# pylint: disable=invalid-name,no-self-use



from __future__ import with_statement
from __future__ import division
from __future__ import absolute_import
import argparse
import json

from flaky import flaky

from allennlp.commands.evaluate import evaluate_from_args, Evaluate
from allennlp.common.testing import AllenNlpTestCase
from io import open


class TestEvaluate(AllenNlpTestCase):
    def setUp(self):
        super(TestEvaluate, self).setUp()

        self.parser = argparse.ArgumentParser(description=u"Testing")
        subparsers = self.parser.add_subparsers(title=u'Commands', metavar=u'')
        Evaluate().add_subparser(u'evaluate', subparsers)

    @flaky
    def test_evaluate_from_args(self):
        kebab_args = [u"evaluate", unicode(self.FIXTURES_ROOT / u"bidaf" / u"serialization" / u"model.tar.gz"),
                      unicode(self.FIXTURES_ROOT / u"data" / u"squad.json"),
                      u"--cuda-device", u"-1"]

        args = self.parser.parse_args(kebab_args)
        metrics = evaluate_from_args(args)
        assert list(metrics.keys()) == set([u'span_acc', u'end_acc', u'start_acc', u'em', u'f1'])

    def test_output_file_evaluate_from_args(self):
        output_file = unicode(self.TEST_DIR / u"metrics.json")
        kebab_args = [u"evaluate", unicode(self.FIXTURES_ROOT / u"bidaf" / u"serialization" / u"model.tar.gz"),
                      unicode(self.FIXTURES_ROOT / u"data" / u"squad.json"),
                      u"--cuda-device", u"-1",
                      u"--output-file", output_file]
        args = self.parser.parse_args(kebab_args)
        computed_metrics = evaluate_from_args(args)
        with open(output_file, u'r') as file:
            saved_metrics = json.load(file)
        assert computed_metrics == saved_metrics





from __future__ import with_statement
from __future__ import absolute_import
#typing
import logging

#overrides
from conllu.parser import parse_line, DEFAULT_FIELDS

from allennlp.common.file_utils import cached_path
from allennlp.data.dataset_readers.dataset_reader import DatasetReader
from allennlp.data.fields import Field, TextField, SequenceLabelField, MetadataField
from allennlp.data.instance import Instance
from allennlp.data.token_indexers import SingleIdTokenIndexer, TokenIndexer
from allennlp.data.tokenizers import Token
from io import open
try:
    from itertools import izip
except:
    izip = zip


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def lazy_parse(text     , fields        = DEFAULT_FIELDS):
    for sentence in text.split(u"\n\n"):
        if sentence:
            yield [parse_line(line, fields)
                   for line in sentence.split(u"\n")
                   if line and not line.strip().startswith(u"#")]


class UniversalDependenciesDatasetReader(DatasetReader):
    u"""
    Reads a file in the conllu Universal Dependencies format. Additionally,
    in order to make it easy to structure a model as predicting arcs, we add a
    dummy 'ROOT_HEAD' token to the start of the sequence.

    Parameters
    ----------
    token_indexers : ``Dict[str, TokenIndexer]``, optional (default=``{"tokens": SingleIdTokenIndexer()}``)
        The token indexers to be applied to the words TextField.
    """
    def __init__(self,
                 token_indexers                          = None,
                 lazy       = False)        :
        super(UniversalDependenciesDatasetReader, self).__init__(lazy)
        self._token_indexers = token_indexers or {u'tokens': SingleIdTokenIndexer()}

    #overrides
    def _read(self, file_path     ):
        # if `file_path` is a URL, redirect to the cache
        file_path = cached_path(file_path)

        with open(file_path, u'r') as conllu_file:
            logger.info(u"Reading UD instances from conllu dataset at: %s", file_path)

            for annotation in  lazy_parse(conllu_file.read()):
                # CoNLLU annotations sometimes add back in words that have been elided
                # in the original sentence; we remove these, as we're just predicting
                # dependencies for the original sentence.
                # We filter by None here as elided words have a non-integer word id,
                # and are replaced with None by the conllu python library.
                annotation = [x for x in annotation if x[u"id"] is not None]

                heads = [x[u"head"] for x in annotation]
                tags = [x[u"deprel"] for x in annotation]
                words = [x[u"form"] for x in annotation]
                pos_tags = [x[u"upostag"] for x in annotation]
                yield self.text_to_instance(words, pos_tags, list(izip(tags, heads)))

    #overrides
    def text_to_instance(self,  # type: ignore
                         words           ,
                         upos_tags           ,
                         dependencies                        = None)            :
        # pylint: disable=arguments-differ
        u"""
        Parameters
        ----------
        words : ``List[str]``, required.
            The words in the sentence to be encoded.
        upos_tags : ``List[str]``, required.
            The universal dependencies POS tags for each word.
        dependencies ``List[Tuple[str, int]]``, optional (default = None)
            A list of  (head tag, head index) tuples. Indices are 1 indexed,
            meaning an index of 0 corresponds to that word being the root of
            the dependency tree.

        Returns
        -------
        An instance containing words, upos tags, dependency head tags and head
        indices as fields.
        """
        fields                   = {}

        tokens = TextField([Token(w) for w in words], self._token_indexers)
        fields[u"words"] = tokens
        fields[u"pos_tags"] = SequenceLabelField(upos_tags, tokens, label_namespace=u"pos")
        if dependencies is not None:
            # We don't want to expand the label namespace with an additional dummy token, so we'll
            # always give the 'ROOT_HEAD' token a label of 'root'.
            fields[u"head_tags"] = SequenceLabelField([x[0] for x in dependencies],
                                                     tokens,
                                                     label_namespace=u"head_tags")
            fields[u"head_indices"] = SequenceLabelField([int(x[1]) for x in dependencies],
                                                        tokens,
                                                        label_namespace=u"head_index_tags")

        fields[u"metadata"] = MetadataField({u"words": words, u"pos": upos_tags})
        return Instance(fields)

UniversalDependenciesDatasetReader = DatasetReader.register(u"universal_dependencies")(UniversalDependenciesDatasetReader)





from __future__ import with_statement
from __future__ import absolute_import
#typing
import json
import logging

#overrides

from allennlp.common.util import JsonDict
from allennlp.data.instance import Instance
from allennlp.data.fields import Field, TextField, ListField, IndexField, LabelField
from allennlp.data.fields import ProductionRuleField, MetadataField
from allennlp.data.tokenizers import Tokenizer, WordTokenizer
from allennlp.data.token_indexers import TokenIndexer, SingleIdTokenIndexer
from allennlp.data.dataset_readers.dataset_reader import DatasetReader
from allennlp.semparse.worlds import NlvrWorld
from io import open


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class NlvrDatasetReader(DatasetReader):
    u"""
    ``DatasetReader`` for the NLVR domain. In addition to the usual methods for reading files and
    instances from text, this class contains a method for creating an agenda of actions that each
    sentence triggers, if needed. Note that we deal with the version of the dataset with structured
    representations of the synthetic images instead of the actual images themselves.

    We support multiple data formats here:
    1) The original json version of the NLVR dataset (http://lic.nlp.cornell.edu/nlvr/) where the
    format of each line in the jsonl file is
    ```
    "sentence": <sentence>,
    "label": <true/false>,
    "identifier": <id>,
    "evals": <dict containing all annotations>,
    "structured_rep": <list of three box representations, where each box is a list of object
    representation dicts, containing fields "x_loc", "y_loc", "color", "type", "size">
    ```

    2) A grouped version (constructed using ``scripts/nlvr/group_nlvr_worlds.py``) where we group
    all the worlds that a sentence appears in. We use the fields ``sentence``, ``label`` and
    ``structured_rep``.  And the format of the grouped files is
    ```
    "sentence": <sentence>,
    "labels": <list of labels corresponding to worlds the sentence appears in>
    "identifier": <id that is only the prefix from the original data>
    "worlds": <list of structured representations>
    ```

    3) A processed version that contains action sequences that lead to the correct denotations (or
    not), using some search. This format is very similar to the grouped format, and has the
    following extra field

    ```
    "correct_sequences": <list of lists of action sequences corresponding to logical forms that
    evaluate to the correct denotations>
    ```

    Parameters
    ----------
    lazy : ``bool`` (optional, default=False)
        Passed to ``DatasetReader``.  If this is ``True``, training will start sooner, but will
        take longer per batch.
    tokenizer : ``Tokenizer`` (optional)
        The tokenizer used for sentences in NLVR. Default is ``WordTokenizer``
    sentence_token_indexers : ``Dict[str, TokenIndexer]`` (optional)
        Token indexers for tokens in input sentences.
        Default is ``{"tokens": SingleIdTokenIndexer()}``
    nonterminal_indexers : ``Dict[str, TokenIndexer]`` (optional)
        Indexers for non-terminals in production rules. The default is to index terminals and
        non-terminals in the same way, but you may want to change it.
        Default is ``{"tokens": SingleIdTokenIndexer("rule_labels")}``
    terminal_indexers : ``Dict[str, TokenIndexer]`` (optional)
        Indexers for terminals in production rules. The default is to index terminals and
        non-terminals in the same way, but you may want to change it.
        Default is ``{"tokens": SingleIdTokenIndexer("rule_labels")}``
    output_agendas : ``bool`` (optional)
        If preparing data for a trainer that uses agendas, set this flag and the datset reader will
        output agendas.
    """
    def __init__(self,
                 lazy       = False,
                 tokenizer            = None,
                 sentence_token_indexers                          = None,
                 nonterminal_indexers                          = None,
                 terminal_indexers                          = None,
                 output_agendas       = True)        :
        super(NlvrDatasetReader, self).__init__(lazy)
        self._tokenizer = tokenizer or WordTokenizer()
        self._sentence_token_indexers = sentence_token_indexers or {u"tokens": SingleIdTokenIndexer()}
        self._nonterminal_indexers = nonterminal_indexers or {u"tokens":
                                                              SingleIdTokenIndexer(u"rule_labels")}
        self._terminal_indexers = terminal_indexers or {u"tokens": SingleIdTokenIndexer(u"rule_labels")}
        self._output_agendas = output_agendas

    #overrides
    def _read(self, file_path     ):
        with open(file_path, u"r") as data_file:
            logger.info(u"Reading instances from lines in file: %s", file_path)
            for line in data_file:
                line = line.strip(u"\n")
                if not line:
                    continue
                data = json.loads(line)
                sentence = data[u"sentence"]
                identifier = data[u"identifier"] if u"identifier" in data else data[u"id"]
                if u"worlds" in data:
                    # This means that we are reading grouped nlvr data. There will be multiple
                    # worlds and corresponding labels per sentence.
                    labels = data[u"labels"]
                    structured_representations = data[u"worlds"]
                else:
                    # We will make lists of labels and structured representations, each with just
                    # one element for consistency.
                    labels = [data[u"label"]]
                    structured_representations = [data[u"structured_rep"]]

                target_sequences                  = None
                # TODO(pradeep): The processed file also has incorrect sequences as well, which are
                # needed if we want to define some sort of a hinge-loss based trainer. Deal with
                # them.
                if u"correct_sequences" in data:
                    # We are reading the processed file and these are the "correct" logical form
                    # sequences. See ``scripts/nlvr/get_nlvr_logical_forms.py``.
                    target_sequences = data[u"correct_sequences"]
                instance = self.text_to_instance(sentence,
                                                 structured_representations,
                                                 labels,
                                                 target_sequences,
                                                 identifier)
                if instance is not None:
                    yield instance

    #overrides
    def text_to_instance(self,  # type: ignore
                         sentence     ,
                         structured_representations                            ,
                         labels            = None,
                         target_sequences                  = None,
                         identifier      = None)            :
        u"""
        Parameters
        ----------
        sentence : ``str``
            The query sentence.
        structured_representations : ``List[List[List[JsonDict]]]``
            A list of Json representations of all the worlds. See expected format in this class' docstring.
        labels : ``List[str]`` (optional)
            List of string representations of the labels (true or false) corresponding to the
            ``structured_representations``. Not required while testing.
        target_sequences : ``List[List[str]]`` (optional)
            List of target action sequences for each element which lead to the correct denotation in
            worlds corresponding to the structured representations.
        identifier : ``str`` (optional)
            The identifier from the dataset if available.
        """
        # pylint: disable=arguments-differ
        worlds = [NlvrWorld(data) for data in structured_representations]
        tokenized_sentence = self._tokenizer.tokenize(sentence)
        sentence_field = TextField(tokenized_sentence, self._sentence_token_indexers)
        production_rule_fields              = []
        instance_action_ids                 = {}
        # TODO(pradeep): Assuming that possible actions are the same in all worlds. This may change
        # later.
        for production_rule in worlds[0].all_possible_actions():
            instance_action_ids[production_rule] = len(instance_action_ids)
            field = ProductionRuleField(production_rule, is_global_rule=True)
            production_rule_fields.append(field)
        action_field = ListField(production_rule_fields)
        worlds_field = ListField([MetadataField(world) for world in worlds])
        fields                   = {u"sentence": sentence_field,
                                    u"worlds": worlds_field,
                                    u"actions": action_field}
        if identifier is not None:
            fields[u"identifier"] = MetadataField(identifier)
        # Depending on the type of supervision used for training the parser, we may want either
        # target action sequences or an agenda in our instance. We check if target sequences are
        # provided, and include them if they are. If not, we'll get an agenda for the sentence, and
        # include that in the instance.
        if target_sequences:
            action_sequence_fields              = []
            for target_sequence in target_sequences:
                index_fields = ListField([IndexField(instance_action_ids[action], action_field)
                                          for action in target_sequence])
                action_sequence_fields.append(index_fields)
                # TODO(pradeep): Define a max length for this field.
            fields[u"target_action_sequences"] = ListField(action_sequence_fields)
        elif self._output_agendas:
            # TODO(pradeep): Assuming every world gives the same agenda for a sentence. This is true
            # now, but may change later too.
            agenda = worlds[0].get_agenda_for_sentence(sentence, add_paths_to_agenda=False)
            assert agenda, u"No agenda found for sentence: %s" % sentence
            # agenda_field contains indices into actions.
            agenda_field = ListField([IndexField(instance_action_ids[action], action_field)
                                      for action in agenda])
            fields[u"agenda"] = agenda_field
        if labels:
            labels_field = ListField([LabelField(label, label_namespace=u'denotations')
                                      for label in labels])
            fields[u"labels"] = labels_field

        return Instance(fields)

NlvrDatasetReader = DatasetReader.register(u"nlvr")(NlvrDatasetReader)



# pylint: disable=no-self-use,invalid-name


from __future__ import division
from __future__ import absolute_import
from allennlp.common.testing import AllenNlpTestCase
from allennlp.models.archival import load_archive
from allennlp.predictors import Predictor


class TestSimpleSeq2SeqPredictor(AllenNlpTestCase):
    def test_uses_named_inputs(self):
        inputs = {
                u"source": u"What kind of test succeeded on its first attempt?",
        }

        archive = load_archive(self.FIXTURES_ROOT / u'encoder_decoder' / u'simple_seq2seq' /
                               u'serialization' / u'model.tar.gz')
        predictor = Predictor.from_archive(archive, u'simple_seq2seq')

        result = predictor.predict_json(inputs)

        predicted_tokens = result.get(u"predicted_tokens")
        assert predicted_tokens is not None
        assert isinstance(predicted_tokens, list)
        assert all(isinstance(x, unicode) for x in predicted_tokens)



# pylint: disable=no-self-use,invalid-name


from __future__ import division
from __future__ import absolute_import
import pytest

from allennlp.data.dataset_readers.conll2003 import Conll2003DatasetReader
from allennlp.common.util import ensure_list
from allennlp.common.testing import AllenNlpTestCase

class TestConll2003Reader(object):
    @pytest.mark.parametrize(u"lazy", (True, False))
    @pytest.mark.parametrize(u"coding_scheme", (u'IOB1', u'BIOUL'))
    def test_read_from_file(self, lazy, coding_scheme):
        conll_reader = Conll2003DatasetReader(lazy=lazy, coding_scheme=coding_scheme)
        instances = conll_reader.read(unicode(AllenNlpTestCase.FIXTURES_ROOT / u'data' / u'conll2003.txt'))
        instances = ensure_list(instances)

        if coding_scheme == u'IOB1':
            expected_labels = [u'I-ORG', u'O', u'I-PER', u'O', u'O', u'I-LOC', u'O']
        else:
            expected_labels = [u'U-ORG', u'O', u'U-PER', u'O', u'O', u'U-LOC', u'O']

        fields = instances[0].fields
        tokens = [t.text for t in fields[u'tokens'].tokens]
        assert tokens == [u'U.N.', u'official', u'Ekeus', u'heads', u'for', u'Baghdad', u'.']
        assert fields[u"tags"].labels == expected_labels

        fields = instances[1].fields
        tokens = [t.text for t in fields[u'tokens'].tokens]
        assert tokens == [u'AI2', u'engineer', u'Joel', u'lives', u'in', u'Seattle', u'.']
        assert fields[u"tags"].labels == expected_labels



# pylint: disable=no-self-use,invalid-name

from __future__ import absolute_import
import torch
import numpy

from allennlp.common.testing import AllenNlpTestCase
from allennlp.modules.seq2seq_encoders import PassThroughEncoder


class TestStackedSelfAttention(AllenNlpTestCase):
    def test_get_dimension_is_correct(self):
        encoder = PassThroughEncoder(input_dim=9)
        assert encoder.get_input_dim() == 9
        assert encoder.get_output_dim() == 9

    def test_pass_through_encoder_passes_through(self):
        encoder = PassThroughEncoder(input_dim=9)
        tensor = torch.randn([2, 3, 9])
        output = encoder(tensor)
        numpy.testing.assert_array_almost_equal(tensor.detach().cpu().numpy(),
                                                output.detach().cpu().numpy())




from __future__ import absolute_import
from allennlp.data.iterators import EpochTrackingBucketIterator
from allennlp.tests.data.iterators.basic_iterator_test import IteratorTest


class EpochTrackingBucketIteratorTest(IteratorTest):
    def setUp(self):
        # The super class creates a self.instances field and populates it with some instances with
        # TextFields.
        super(EpochTrackingBucketIteratorTest, self).setUp()
        self.iterator = EpochTrackingBucketIterator(sorting_keys=[[u"text", u"num_tokens"]])
        # We'll add more to create a second dataset.
        self.more_instances = [
                self.create_instance([u"this", u"is", u"a", u"sentence"]),
                self.create_instance([u"this", u"is", u"in", u"the", u"second", u"dataset"]),
                self.create_instance([u"so", u"is", u"this", u"one"])
                ]

    def test_iterator_tracks_epochs_per_dataset(self):
        generated_dataset1 = list(self.iterator(self.instances, num_epochs=2))
        generated_dataset2 = list(self.iterator(self.more_instances, num_epochs=2))

        # First dataset has five sentences. See ``IteratorTest.setUp``
        assert generated_dataset1[0][u"epoch_num"] == [0, 0, 0, 0, 0]
        assert generated_dataset1[1][u"epoch_num"] == [1, 1, 1, 1, 1]
        # Second dataset has three sentences.
        assert generated_dataset2[0][u"epoch_num"] == [0, 0, 0]
        assert generated_dataset2[1][u"epoch_num"] == [1, 1, 1]



# pylint: disable=no-self-use,invalid-name,protected-access

from __future__ import absolute_import
import torch

from allennlp.common.testing import AllenNlpTestCase
from allennlp.training.metrics import ConllCorefScores

class ConllCorefScoresTest(AllenNlpTestCase):
    def test_get_predicted_clusters(self):
        top_spans = torch.Tensor([[0, 1], [4, 6], [8, 9]]).long()
        antecedent_indices = torch.Tensor([[-1, -1, -1],
                                           [0, -1, -1],
                                           [0, 1, -1]]).long()
        predicted_antecedents = torch.Tensor([-1, -1, 1]).long()
        clusters, mention_to_cluster = ConllCorefScores.get_predicted_clusters(top_spans,
                                                                               antecedent_indices,
                                                                               predicted_antecedents)
        assert len(clusters) == 1
        assert set(clusters[0]) == set([(4, 6), (8, 9)])
        assert mention_to_cluster == {(4, 6): clusters[0], (8, 9): clusters[0]}




from __future__ import absolute_import
#overrides

from nltk.sem.logic import TRUTH_TYPE

from allennlp.common.testing import AllenNlpTestCase
from allennlp.semparse.worlds.world import World
from allennlp.semparse import ActionSpaceWalker


class FakeWorldWithAssertions(World):
    # pylint: disable=abstract-method
    #overrides
    def get_valid_starting_types(self):
        return set([TRUTH_TYPE])

    #overrides
    def get_valid_actions(self):
        # This grammar produces true/false statements like
        # (object_exists all_objects)
        # (object_exists (black all_objects))
        # (object_exists (triangle all_objects))
        # (object_exists (touch_wall all_objects))
        # (object_exists (triangle (black all_objects)))
        # ...
        actions = {u'@start@': [u'@start@ -> t'],
                   u't': [u't -> [<o,t>, o]'],
                   u'<o,t>': [u'<o,t> -> object_exists'],
                   u'o': [u'o -> [<o,o>, o]', u'o -> all_objects'],
                   u'<o,o>': [u'<o,o> -> black', u'<o,o> -> triangle', u'<o,o> -> touch_wall']}
        return actions

    #overrides
    def is_terminal(self, symbol     )        :
        return symbol in set([u'object_exists', u'all_objects', u'black', u'triangle', u'touch_wall'])


class ActionSpaceWalkerTest(AllenNlpTestCase):
    def setUp(self):
        super(ActionSpaceWalkerTest, self).setUp()
        self.world = FakeWorldWithAssertions()
        self.walker = ActionSpaceWalker(self.world, max_path_length=10)

    def test_get_logical_forms_with_agenda(self):
        black_logical_forms = self.walker.get_logical_forms_with_agenda([u'<o,o> -> black'])
        # These are all the possible logical forms with black
        assert len(black_logical_forms) == 25
        shortest_logical_form = self.walker.get_logical_forms_with_agenda([u'<o,o> -> black'], 1)[0]
        # This is the shortest complete logical form with black
        assert shortest_logical_form == u'(object_exists (black all_objects))'
        black_triangle_touch_forms = self.walker.get_logical_forms_with_agenda([u'<o,o> -> black',
                                                                                u'<o,o> -> triangle',
                                                                                u'<o,o> -> touch_wall'])
        # Permutations of the three functions. There will not be repetitions of any functions
        # because we limit the length of paths to 10 above.
        assert set(black_triangle_touch_forms) == set([
                u'(object_exists (black (triangle (touch_wall all_objects))))',
                u'(object_exists (black (touch_wall (triangle all_objects))))',
                u'(object_exists (triangle (black (touch_wall all_objects))))',
                u'(object_exists (triangle (touch_wall (black all_objects))))',
                u'(object_exists (touch_wall (black (triangle all_objects))))',
                u'(object_exists (touch_wall (triangle (black all_objects))))'])

    def test_get_all_logical_forms(self):
        # get_all_logical_forms should sort logical forms by length.
        ten_shortest_logical_forms = self.walker.get_all_logical_forms(max_num_logical_forms=10)
        shortest_logical_form = ten_shortest_logical_forms[0]
        assert shortest_logical_form == u'(object_exists all_objects)'
        length_three_logical_forms = ten_shortest_logical_forms[1:4]
        assert set(length_three_logical_forms) == set([u'(object_exists (black all_objects))',
                                                   u'(object_exists (touch_wall all_objects))',
                                                   u'(object_exists (triangle all_objects))'])




from __future__ import absolute_import
from allennlp.modules.encoder_base import _EncoderBase
from allennlp.common import Registrable


class Seq2SeqEncoder(_EncoderBase, Registrable):
    u"""
    A ``Seq2SeqEncoder`` is a ``Module`` that takes as input a sequence of vectors and returns a
    modified sequence of vectors.  Input shape: ``(batch_size, sequence_length, input_dim)``; output
    shape: ``(batch_size, sequence_length, output_dim)``.

    We add two methods to the basic ``Module`` API: :func:`get_input_dim()` and :func:`get_output_dim()`.
    You might need this if you want to construct a ``Linear`` layer using the output of this encoder,
    or to raise sensible errors for mis-matching input dimensions.
    """
    def get_input_dim(self)       :
        u"""
        Returns the dimension of the vector input for each element in the sequence input
        to a ``Seq2SeqEncoder``. This is `not` the shape of the input tensor, but the
        last element of that shape.
        """
        raise NotImplementedError

    def get_output_dim(self)       :
        u"""
        Returns the dimension of each vector in the sequence output by this ``Seq2SeqEncoder``.
        This is `not` the shape of the returned tensor, but the last element of that shape.
        """
        raise NotImplementedError

    def is_bidirectional(self)        :
        u"""
        Returns ``True`` if this encoder is bidirectional.  If so, we assume the forward direction
        of the encoder is the first half of the final dimension, and the backward direction is the
        second half.
        """
        raise NotImplementedError




from __future__ import absolute_import
#typing
import itertools

#overrides

from allennlp.common.checks import ConfigurationError
from allennlp.common.util import pad_sequence_to_length
from allennlp.data.tokenizers.token import Token
from allennlp.data.token_indexers.token_indexer import TokenIndexer
from allennlp.data.vocabulary import Vocabulary
from allennlp.data.tokenizers.character_tokenizer import CharacterTokenizer
try:
    from itertools import izip
except:
    izip = zip



class TokenCharactersIndexer(TokenIndexer):
    u"""
    This :class:`TokenIndexer` represents tokens as lists of character indices.

    Parameters
    ----------
    namespace : ``str``, optional (default=``token_characters``)
        We will use this namespace in the :class:`Vocabulary` to map the characters in each token
        to indices.
    character_tokenizer : ``CharacterTokenizer``, optional (default=``CharacterTokenizer()``)
        We use a :class:`CharacterTokenizer` to handle splitting tokens into characters, as it has
        options for byte encoding and other things.  The default here is to instantiate a
        ``CharacterTokenizer`` with its default parameters, which uses unicode characters and
        retains casing.
    """
    # pylint: disable=no-self-use
    def __init__(self,
                 namespace      = u'token_characters',
                 character_tokenizer                     = CharacterTokenizer())        :
        self._namespace = namespace
        self._character_tokenizer = character_tokenizer

    #overrides
    def count_vocab_items(self, token       , counter                           ):
        if token.text is None:
            raise ConfigurationError(u'TokenCharactersIndexer needs a tokenizer that retains text')
        for character in self._character_tokenizer.tokenize(token.text):
            # If `text_id` is set on the character token (e.g., if we're using byte encoding), we
            # will not be using the vocab for this character.
            if getattr(character, u'text_id', None) is None:
                counter[self._namespace][character.text] += 1

    #overrides
    def tokens_to_indices(self,
                          tokens             ,
                          vocabulary            ,
                          index_name     )                              :
        indices                  = []
        for token in tokens:
            token_indices            = []
            if token.text is None:
                raise ConfigurationError(u'TokenCharactersIndexer needs a tokenizer that retains text')
            for character in self._character_tokenizer.tokenize(token.text):
                if getattr(character, u'text_id', None) is not None:
                    # `text_id` being set on the token means that we aren't using the vocab, we just
                    # use this id instead.
                    index = character.text_id
                else:
                    index = vocabulary.get_token_index(character.text, self._namespace)
                token_indices.append(index)
            indices.append(token_indices)
        return {index_name: indices}

    #overrides
    def get_padding_lengths(self, token           )                  :
        return {u'num_token_characters': len(token)}

    #overrides
    def get_padding_token(self)             :
        return []

    #overrides
    def pad_token_sequence(self,
                           tokens                            ,
                           desired_num_tokens                ,
                           padding_lengths                )                              :
        # Pad the tokens.
        # tokens has only one key...
        key = list(tokens.keys())[0]

        padded_tokens = pad_sequence_to_length(
                tokens[key], desired_num_tokens[key],
                default_value=self.get_padding_token
        )

        # Pad the characters within the tokens.
        desired_token_length = padding_lengths[u'num_token_characters']
        longest_token            = max(tokens[key], key=len, default=[])
        padding_value = 0
        if desired_token_length > len(longest_token):
            # Since we want to pad to greater than the longest token, we add a
            # "dummy token" so we can take advantage of the fast implementation of itertools.zip_longest.
            padded_tokens.append([padding_value] * desired_token_length)
        # pad the list of lists to the longest sublist, appending 0's
        padded_tokens = list(izip(*itertools.zip_longest(*padded_tokens, fillvalue=padding_value)))
        if desired_token_length > len(longest_token):
            # Removes the "dummy token".
            padded_tokens.pop()
        # Truncates all the tokens to the desired length, and return the result.
        return {key: [list(token[:desired_token_length]) for token in padded_tokens]}

TokenCharactersIndexer = TokenIndexer.register(u"characters")(TokenCharactersIndexer)




from __future__ import absolute_import
#typing

#overrides
import torch

from allennlp.nn import util

class SpanPruner(torch.nn.Module):
    u"""
    This module scores and prunes span-based representations using a parameterised scoring
    function and a threshold.

    Parameters
    ----------
    scorer : ``torch.nn.Module``, required.
        A module which, given a tensor of shape (batch_size, num_spans, embedding_size),
        produces a tensor of shape (batch_size, num_spans, 1), representing a scalar score
        per span in the tensor.
    """
    def __init__(self, scorer                 )        :
        super(SpanPruner, self).__init__()
        self._scorer = scorer

    #overrides
    def forward(self, # pylint: disable=arguments-differ
                span_embeddings                   ,
                span_mask                  ,
                num_spans_to_keep     ):                                              
                                                                                     
        u"""
        Extracts the top-k scoring spans with respect to the scorer. We additionally return
        the indices of the top-k in their original order, not ordered by score, so that we
        can rely on the ordering to consider the previous k spans as antecedents for each
        span later.

        Parameters
        ----------
        span_embeddings : ``torch.FloatTensor``, required.
            A tensor of shape (batch_size, num_spans, embedding_size), representing
            the set of embedded span representations.
        span_mask : ``torch.LongTensor``, required.
            A tensor of shape (batch_size, num_spans), denoting unpadded elements
            of ``span_embeddings``.
        num_spans_to_keep : ``int``, required.
            The number of spans to keep when pruning.

        Returns
        -------
        top_span_embeddings : ``torch.FloatTensor``
            The span representations of the top-k scoring spans.
            Has shape (batch_size, num_spans_to_keep, embedding_size).
        top_span_mask : ``torch.LongTensor``
            The coresponding mask for ``top_span_embeddings``.
            Has shape (batch_size, num_spans_to_keep).
        top_span_indices : ``torch.IntTensor``
            The indices of the top-k scoring spans into the original ``span_embeddings``
            tensor. This is returned because it can be useful to retain pointers to
            the original spans, if each span is being scored by multiple distinct
            scorers, for instance. Has shape (batch_size, num_spans_to_keep).
        top_span_scores : ``torch.FloatTensor``
            The values of the top-k scoring spans.
            Has shape (batch_size, num_spans_to_keep, 1).
        """
        span_mask = span_mask.unsqueeze(-1)
        num_spans = span_embeddings.size(1)
        # Shape: (batch_size, num_spans, 1)
        span_scores = self._scorer(span_embeddings)

        if span_scores.size(-1) != 1 or span_scores.dim() != 3:
            raise ValueError("The scorer passed to SpanPruner must produce a tensor of shape"
                             "(batch_size, num_spans, 1), but found shape {span_scores.size()}")
        # Make sure that we don't select any masked spans by
        # setting their scores to be -inf.
        span_scores += span_mask.log()

        # Shape: (batch_size, num_spans_to_keep, 1)
        _, top_span_indices = span_scores.topk(num_spans_to_keep, 1)

        # Now we order the selected indices in increasing order with
        # respect to their indices (and hence, with respect to the
        # order they originally appeared in the ``span_embeddings`` tensor).
        top_span_indices, _ = torch.sort(top_span_indices, 1)

        # Shape: (batch_size, num_spans_to_keep)
        top_span_indices = top_span_indices.squeeze(-1)

        # Shape: (batch_size * num_spans_to_keep)
        # torch.index_select only accepts 1D indices, but here
        # we need to select spans for each element in the batch.
        flat_top_span_indices = util.flatten_and_batch_shift_indices(top_span_indices, num_spans)

        # Shape: (batch_size, num_spans_to_keep, embedding_size)
        top_span_embeddings = util.batched_index_select(span_embeddings,
                                                        top_span_indices,
                                                        flat_top_span_indices)
        # Shape: (batch_size, num_spans_to_keep)
        top_span_mask = util.batched_index_select(span_mask,
                                                  top_span_indices,
                                                  flat_top_span_indices)

        # Shape: (batch_size, num_spans_to_keep, 1)
        top_span_scores = util.batched_index_select(span_scores,
                                                    top_span_indices,
                                                    flat_top_span_indices)

        return top_span_embeddings, top_span_mask.squeeze(-1), top_span_indices, top_span_scores





from __future__ import division
from __future__ import absolute_import
import os

from allennlp.common.testing import AllenNlpTestCase
from allennlp.models.archival import load_archive
from allennlp.predictors import Predictor


class TestNlvrParserPredictor(AllenNlpTestCase):
    def setUp(self):
        super(TestNlvrParserPredictor, self).setUp()
        self.inputs = {u'worlds': [[[{u'y_loc': 80, u'type': u'triangle', u'color': u'#0099ff', u'x_loc': 80,
                                     u'size': 20}],
                                   [{u'y_loc': 80, u'type': u'square', u'color': u'Yellow', u'x_loc': 13,
                                     u'size': 20}],
                                   [{u'y_loc': 67, u'type': u'triangle', u'color': u'Yellow', u'x_loc': 35,
                                     u'size': 10}]],
                                  [[{u'y_loc': 8, u'type': u'square', u'color': u'Yellow', u'x_loc': 57,
                                     u'size': 30}],
                                   [{u'y_loc': 43, u'type': u'square', u'color': u'#0099ff', u'x_loc': 70,
                                     u'size': 30}],
                                   [{u'y_loc': 59, u'type': u'square', u'color': u'Yellow', u'x_loc': 47,
                                     u'size': 10}]]],
                       u'identifier': u'fake_id',
                       u'sentence': u'Each grey box contains atleast one yellow object touching the edge'}

    def test_predictor_with_coverage_parser(self):
        archive_dir = self.FIXTURES_ROOT / u'semantic_parsing' / u'nlvr_coverage_semantic_parser' / u'serialization'
        archive = load_archive(os.path.join(archive_dir, u'model.tar.gz'))
        predictor = Predictor.from_archive(archive, u'nlvr-parser')

        result = predictor.predict_json(self.inputs)
        assert u'logical_form' in result
        assert u'denotations' in result
        # result['denotations'] is a list corresponding to k-best logical forms, where k is 1 by
        # default.
        assert len(result[u'denotations'][0]) == 2  # Because there are two worlds in the input.

    def test_predictor_with_direct_parser(self):
        archive_dir = self.FIXTURES_ROOT / u'semantic_parsing' / u'nlvr_direct_semantic_parser' / u'serialization'
        archive = load_archive(os.path.join(archive_dir, u'model.tar.gz'))
        predictor = Predictor.from_archive(archive, u'nlvr-parser')

        result = predictor.predict_json(self.inputs)
        assert u'logical_form' in result
        assert u'denotations' in result
        # result['denotations'] is a list corresponding to k-best logical forms, where k is 1 by
        # default.
        assert len(result[u'denotations'][0]) == 2  # Because there are two worlds in the input.




from __future__ import absolute_import
#typing
from datetime import datetime
import re

from collections import defaultdict
from allennlp.data.tokenizers import Token
try:
    from itertools import izip
except:
    izip = zip


TWELVE_TO_TWENTY_FOUR = 1200
HOUR_TO_TWENTY_FOUR = 100
HOURS_IN_DAY = 2400
AROUND_RANGE = 30

APPROX_WORDS = [u'about', u'around', u'approximately']
WORDS_PRECEDING_TIME = [u'at', u'between', u'to', u'before', u'after']

def get_times_from_utterance(utterance     ,
                             char_offset_to_token_index                ,
                             indices_of_approximate_words          )                        :
    u"""
    Given an utterance, we get the numbers that correspond to times and convert them to
    values that may appear in the query. For example: convert ``7pm`` to ``1900``.
    """

    pm_linking_dict = _time_regex_match(ur'\d+pm',
                                        utterance,
                                        char_offset_to_token_index,
                                        lambda match: [int(match.rstrip(u'pm'))
                                                       * HOUR_TO_TWENTY_FOUR +
                                                       TWELVE_TO_TWENTY_FOUR],
                                        indices_of_approximate_words)

    am_linking_dict = _time_regex_match(ur'\d+am',
                                        utterance,
                                        char_offset_to_token_index,
                                        lambda match: [int(match.rstrip(u'am'))
                                                       * HOUR_TO_TWENTY_FOUR],
                                        indices_of_approximate_words)

    oclock_linking_dict = _time_regex_match(ur"\d+\so'clock",
                                            utterance,
                                            char_offset_to_token_index,
                                            lambda match: digit_to_query_time(match.rstrip(u"o'clock")),
                                            indices_of_approximate_words)

    times_linking_dict                       = defaultdict(list)
    linking_dicts = [pm_linking_dict, am_linking_dict, oclock_linking_dict]

    for linking_dict in linking_dicts:
        for key, value in list(linking_dict.items()):
            times_linking_dict[key].extend(value)

    return times_linking_dict

def get_date_from_utterance(tokenized_utterance             ,
                            year      = 1993,
                            month      = None,
                            day      = None)            :
    u"""
    When the year is not explicitly mentioned in the utterance, the query assumes that
    it is 1993 so we do the same here. If there is no mention of the month or day then
    we do not return any dates from the utterance.
    """
    utterance = u' '.join([token.text for token in tokenized_utterance])
    year_result = re.findall(ur'199[0-4]', utterance)
    if year_result:
        year = int(year_result[0])

    for token in tokenized_utterance:
        if token.text in MONTH_NUMBERS:
            month = MONTH_NUMBERS[token.text]
        if token.text in DAY_NUMBERS:
            day = DAY_NUMBERS[token.text]

    for tens, digits in izip(tokenized_utterance, tokenized_utterance[1:]):
        bigram = u' '.join([tens.text, digits.text])
        if bigram in DAY_NUMBERS:
            day = DAY_NUMBERS[bigram]
    if month and day:
        return datetime(year, month, day)
    return None

def get_numbers_from_utterance(utterance     , tokenized_utterance             )                        :
    u"""
    Given an utterance, this function finds all the numbers that are in the action space. Since we need to
    keep track of linking scores, we represent the numbers as a dictionary, where the keys are the string
    representation of the number and the values are lists of the token indices that triggers that number.
    """
    # When we use a regex to find numbers or strings, we need a mapping from
    # the character to which token triggered it.
    char_offset_to_token_index = dict((token.idx, token_index)
                                  for token_index, token in enumerate(tokenized_utterance))

    # We want to look up later for each time whether it appears after a word
    # such as "about" or "approximately".
    indices_of_approximate_words = set(index for index, token in enumerate(tokenized_utterance)
                                    if token.text in APPROX_WORDS)

    indices_of_words_preceding_time = set(index for index, token in enumerate(tokenized_utterance)
                                       if token.text in WORDS_PRECEDING_TIME)

    number_linking_dict                       = defaultdict(list)
    for token_index, token in enumerate(tokenized_utterance):
        if token.text.isdigit():
            if token_index - 1 in indices_of_words_preceding_time:
                for time in digit_to_query_time(token.text):
                    number_linking_dict[unicode(time)].append(token_index)
            else:
                number_linking_dict[token.text].append(token_index)

    times_linking_dict = get_times_from_utterance(utterance,
                                                  char_offset_to_token_index,
                                                  indices_of_approximate_words)

    for key, value in list(times_linking_dict.items()):
        number_linking_dict[key].extend(value)

    for index, token in enumerate(tokenized_utterance):
        for number in NUMBER_TRIGGER_DICT.get(token.text, []):
            number_linking_dict[number].append(index)

    for tens, digits in izip(tokenized_utterance, tokenized_utterance[1:]):
        bigram = u' '.join([tens.text, digits.text])
        if bigram in DAY_NUMBERS:
            number_linking_dict[unicode(DAY_NUMBERS[bigram])].append(len(tokenized_utterance) - 1)

    return number_linking_dict

def digit_to_query_time(digit     )             :
    u"""
    Given a digit in the utterance, return a list of the times that it corresponds to.
    """
    if int(digit) % 12 == 0:
        return [0, 1200, 2400]
    return [int(digit) * HOUR_TO_TWENTY_FOUR,
            (int(digit) * HOUR_TO_TWENTY_FOUR + TWELVE_TO_TWENTY_FOUR) % HOURS_IN_DAY]

def get_approximate_times(times           )             :
    u"""
    Given a list of times that follow a word such as ``about``,
    we return a list of times that could appear in the query as a result
    of this. For example if ``about 7pm`` appears in the utterance, then
    we also want to add ``1830`` and ``1930``.
    """
    approximate_times = []
    for time in times:
        approximate_times.append((time + AROUND_RANGE) % HOURS_IN_DAY)
        # The number system is not base 10 here, there are 60 minutes
        # in an hour, so we can't simply add time - AROUND_RANGE.
        approximate_times.append((time - HOUR_TO_TWENTY_FOUR + AROUND_RANGE) % HOURS_IN_DAY)
    return approximate_times

def _time_regex_match(regex     ,
                      utterance     ,
                      char_offset_to_token_index                ,
                      map_match_to_query_value                            ,
                      indices_of_approximate_words          )                        :
    ur"""
    Given a regex for matching times in the utterance, we want to convert the matches
    to the values that appear in the query and token indices they correspond to.

    ``char_offset_to_token_index`` is a dictionary that maps from the character offset to
    the token index, we use this to look up what token a regex match corresponds to.
    ``indices_of_approximate_words`` are the token indices of the words such as ``about`` or
    ``approximately``. We use this to check if a regex match is preceded by one of these words.
    If it is, we also want to add the times that define this approximate time range.

    ``map_match_to_query_value`` is a function that converts the regex matches to the
    values that appear in the query. For example, we may pass in a regex such as ``\d+pm``
    that matches times such as ``7pm``. ``map_match_to_query_value`` would be a function that
    takes ``7pm`` as input and returns ``1900``.
    """
    linking_scores_dict                       = defaultdict(list)
    number_regex = re.compile(regex)
    for match in number_regex.finditer(utterance):
        query_values = map_match_to_query_value(match.group())
        # If the time appears after a word like ``about`` then we also add
        # the times that mark the start and end of the allowed range.
        approximate_times = []
        if char_offset_to_token_index.get(match.start(), 0) - 1 in indices_of_approximate_words:
            approximate_times.extend(get_approximate_times(query_values))
        query_values.extend(approximate_times)
        if match.start() in char_offset_to_token_index:
            for query_value in query_values:
                linking_scores_dict[unicode(query_value)].append(char_offset_to_token_index[match.start()])
    return linking_scores_dict

def get_trigger_dict(trigger_lists                 ,
                     trigger_dicts                            )                        :
    merged_trigger_dict                       = defaultdict(list)
    for trigger_list in trigger_lists:
        for trigger in trigger_list:
            merged_trigger_dict[trigger.lower()].append(trigger)

    for trigger_dict in trigger_dicts:
        for key, value in list(trigger_dict.items()):
            merged_trigger_dict[key.lower()].extend(value)

    return merged_trigger_dict

def convert_to_string_list_value_dict(trigger_dict                )                        :
    return dict((key, [unicode(value)]) for key, value in list(trigger_dict.items()))

AIRLINE_CODES = {u'alaska': [u'AS'],
                 u'alliance': [u'3J'],
                 u'alpha': [u'7V'],
                 u'america west': [u'HP'],
                 u'american': [u'AA'],
                 u'american trans': [u'TZ'],
                 u'argentina': [u'AR'],
                 u'atlantic': [u'DH'],
                 u'atlantic.': [u'EV'],
                 u'braniff.': [u'BE'],
                 u'british': [u'BA'],
                 u'business': [u'HQ'],
                 u'canada': [u'AC'],
                 u'canadian': [u'CP'],
                 u'carnival': [u'KW'],
                 u'christman': [u'SX'],
                 u'colgan': [u'9L'],
                 u'comair': [u'OH'],
                 u'continental': [u'CO'],
                 u'czecho': [u'OK'],
                 u'delta': [u'DL'],
                 u'eastern': [u'EA'],
                 u'express': [u'9E'],
                 u'grand': [u'QD'],
                 u'lufthansa': [u'LH'],
                 u'mesaba': [u'XJ'],
                 u'mgm': [u'MG'],
                 u'midwest': [u'YX'],
                 u'nation': [u'NX'],
                 u'northeast': [u'2V'],
                 u'northwest': [u'NW'],
                 u'ontario': [u'GX'],
                 u'ontario express': [u'9X'],
                 u'precision': [u'RP'],
                 u'royal': [u'AT'],
                 u'sabena': [u'SN'],
                 u'sky': [u'OO'],
                 u'south': [u'WN'],
                 u'states': [u'9N'],
                 u'thai': [u'TG'],
                 u'tower': [u'FF'],
                 u'twa': [u'TW'],
                 u'united': [u'UA'],
                 u'us': [u'US'],
                 u'west': [u'OE'],
                 u'wisconson': [u'ZW'],
                 u'world': [u'RZ']}

CITY_CODES = {u'ATLANTA': [u'MATL'],
              u'BALTIMORE': [u'BBWI'],
              u'BOSTON': [u'BBOS'],
              u'BURBANK': [u'BBUR'],
              u'CHARLOTTE': [u'CCLT'],
              u'CHICAGO': [u'CCHI'],
              u'CINCINNATI': [u'CCVG'],
              u'CLEVELAND': [u'CCLE'],
              u'COLUMBUS': [u'CCMH'],
              u'DALLAS': [u'DDFW'],
              u'DENVER': [u'DDEN'],
              u'DETROIT': [u'DDTT'],
              u'FORT WORTH': [u'FDFW'],
              u'HOUSTON': [u'HHOU'],
              u'KANSAS CITY': [u'MMKC'],
              u'LAS VEGAS': [u'LLAS'],
              u'LONG BEACH': [u'LLGB'],
              u'LOS ANGELES': [u'LLAX'],
              u'MEMPHIS': [u'MMEM'],
              u'MIAMI': [u'MMIA'],
              u'MILWAUKEE': [u'MMKE'],
              u'MINNEAPOLIS': [u'MMSP'],
              u'MONTREAL': [u'YYMQ'],
              u'NASHVILLE': [u'BBNA'],
              u'NEW YORK': [u'NNYC'],
              u'NEWARK': [u'JNYC'],
              u'OAKLAND': [u'OOAK'],
              u'ONTARIO': [u'OONT'],
              u'ORLANDO': [u'OORL'],
              u'PHILADELPHIA': [u'PPHL'],
              u'PHOENIX': [u'PPHX'],
              u'PITTSBURGH': [u'PPIT'],
              u'SALT LAKE CITY': [u'SSLC'],
              u'SAN DIEGO': [u'SSAN'],
              u'SAN FRANCISCO': [u'SSFO'],
              u'SAN JOSE': [u'SSJC'],
              u'SEATTLE': [u'SSEA'],
              u'ST. LOUIS': [u'SSTL'],
              u'ST. PAUL': [u'SMSP'],
              u'ST. PETERSBURG': [u'STPA'],
              u'TACOMA': [u'TSEA'],
              u'TAMPA': [u'TTPA'],
              u'TORONTO': [u'YYTO'],
              u'WASHINGTON': [u'WWAS'],
              u'WESTCHESTER COUNTY': [u'HHPN']}

MONTH_NUMBERS = {u'january': 1,
                 u'february': 2,
                 u'march': 3,
                 u'april': 4,
                 u'may': 5,
                 u'june': 6,
                 u'july': 7,
                 u'august': 8,
                 u'september': 9,
                 u'october': 10,
                 u'november': 11,
                 u'december': 12}

DAY_NUMBERS = {u'first': 1,
               u'second': 2,
               u'third': 3,
               u'fourth': 4,
               u'fifth': 5,
               u'sixth': 6,
               u'seventh': 7,
               u'eighth': 8,
               u'ninth': 9,
               u'tenth': 10,
               u'eleventh': 11,
               u'twelfth': 12,
               u'thirteenth': 13,
               u'fourteenth': 14,
               u'fifteenth': 15,
               u'sixteenth': 16,
               u'seventeenth': 17,
               u'eighteenth': 18,
               u'nineteenth': 19,
               u'twentieth': 20,
               u'twenty first': 21,
               u'twenty second': 22,
               u'twenty third': 23,
               u'twenty fourth': 24,
               u'twenty fifth': 25,
               u'twenty sixth': 26,
               u'twenty seventh': 27,
               u'twenty eighth': 28,
               u'twenty ninth': 29,
               u'thirtieth': 30,
               u'thirty first': 31}

GROUND_SERVICE = {u'air taxi': [u'AIR TAXI OPERATION'],
                  u'car': [u'RENTAL CAR'],
                  u'limo': [u'LIMOUSINE'],
                  u'rapid': [u'RAPID TRANSIT'],
                  u'rental': [u'RENTAL CAR'],
                  u'taxi': [u'TAXI']}

MISC_STR = {u"every day" : [u"DAILY"]}

MISC_TIME_TRIGGERS = {u'morning': [u'0', u'1200'],
                      u'afternoon': [u'1200', u'1800'],
                      u'after': [u'1200', u'1800'],
                      u'evening': [u'1800', u'2200'],
                      u'late evening': [u'2000', u'2200'],
                      u'lunch': [u'1400'],
                      u'noon': [u'1200']}

TABLES = {u'aircraft': [u'aircraft_code', u'aircraft_description',
                       u'manufacturer', u'basic_type', u'propulsion',
                       u'wide_body', u'pressurized'],
          u'airline': [u'airline_name', u'airline_code'],
          u'airport': [u'airport_code', u'airport_name', u'airport_location',
                      u'state_code', u'country_name', u'time_zone_code',
                      u'minimum_connect_time'],
          u'airport_service': [u'city_code', u'airport_code', u'miles_distant',
                              u'direction', u'minutes_distant'],
          u'city': [u'city_code', u'city_name', u'state_code', u'country_name', u'time_zone_code'],
          u'class_of_service': [u'booking_class', u'rank', u'class_description'],
          u'date_day': [u'month_number', u'day_number', u'year', u'day_name'],
          u'days': [u'days_code', u'day_name'],
          u'equipment_sequence': [u'aircraft_code_sequence', u'aircraft_code'],
          u'fare': [u'fare_id', u'from_airport', u'to_airport', u'fare_basis_code',
                   u'fare_airline', u'restriction_code', u'one_direction_cost',
                   u'round_trip_cost', u'round_trip_required'],
          u'fare_basis': [u'fare_basis_code', u'booking_class', u'class_type', u'premium', u'economy',
                         u'discounted', u'night', u'season', u'basis_days'],
          u'flight': [u'flight_id', u'flight_days', u'from_airport', u'to_airport', u'departure_time',
                     u'arrival_time', u'airline_flight', u'airline_code', u'flight_number',
                     u'aircraft_code_sequence', u'meal_code', u'stops', u'connections',
                     u'dual_carrier', u'time_elapsed'],
          u'flight_fare': [u'flight_id', u'fare_id'],
          u'flight_leg': [u'flight_id', u'leg_number', u'leg_flight'],
          u'flight_stop': [u'flight_id', u'stop_number', u'stop_days', u'stop_airport',
                          u'arrival_time', u'arrival_airline', u'arrival_flight_number',
                          u'departure_time', u'departure_airline', u'departure_flight_number',
                          u'stop_time'],
          u'food_service': [u'meal_code', u'meal_number', u'compartment', u'meal_description'],
          u'ground_service': [u'city_code', u'airport_code', u'transport_type', u'ground_fare'],
          u'month': [u'month_number', u'month_name'],
          u'restriction': [u'restriction_code', u'advance_purchase', u'stopovers',
                          u'saturday_stay_required', u'minimum_stay', u'maximum_stay',
                          u'application', u'no_discounts'],
          u'state': [u'state_code', u'state_name', u'country_name']}

DAY_OF_WEEK_DICT = {u'weekdays' : [u'MONDAY', u'TUESDAY', u'WEDNESDAY', u'THURSDAY', u'FRIDAY']}

YES_NO = {u'one way': [u'NO'],
          u'economy': [u'YES']}

CITY_AIRPORT_CODES = {u'atlanta' : [u'ATL'],
                      u'boston' : [u'BOS'],
                      u'baltimore': [u'BWI'],
                      u'charlotte': [u'CLT'],
                      u'dallas': [u'DFW'],
                      u'detroit': [u'DTW'],
                      u'la guardia': [u'LGA'],
                      u'oakland': [u'OAK'],
                      u'philadelphia': [u'PHL'],
                      u'pittsburgh': [u'PIT'],
                      u'san francisco': [u'SFO'],
                      u'toronto': [u'YYZ']}

AIRPORT_CODES = [u'ATL', u'NA', u'OS', u'UR', u'WI', u'CLE', u'CLT', u'CMH',
                 u'CVG', u'DAL', u'DCA', u'DEN', u'DET', u'DFW', u'DTW',
                 u'EWR', u'HOU', u'HPN', u'IAD', u'IAH', u'IND', u'JFK',
                 u'LAS', u'LAX', u'LGA', u'LG', u'MCI', u'MCO', u'MDW', u'MEM',
                 u'MIA', u'MKE', u'MSP', u'OAK', u'ONT', u'ORD', u'PHL', u'PHX',
                 u'PIE', u'PIT', u'SAN', u'SEA', u'SFO', u'SJC', u'SLC',
                 u'STL', u'TPA', u'YKZ', u'YMX', u'YTZ', u'YUL', u'YYZ']

AIRLINE_CODE_LIST = [u'AR', u'3J', u'AC', u'9X', u'ZW', u'AS', u'7V',
                     u'AA', u'TZ', u'HP', u'DH', u'EV', u'BE', u'BA',
                     u'HQ', u'CP', u'KW', u'SX', u'9L', u'OH', u'CO',
                     u'OK', u'DL', u'9E', u'QD', u'LH', u'XJ', u'MG',
                     u'YX', u'NX', u'2V', u'NW', u'RP', u'AT', u'SN',
                     u'OO', u'WN', u'TG', u'FF', u'9N', u'TW', u'RZ',
                     u'UA', u'US', u'OE']

CITIES = [u'NASHVILLE', u'BOSTON', u'BURBANK', u'BALTIMORE', u'CHICAGO', u'CLEVELAND',
          u'CHARLOTTE', u'COLUMBUS', u'CINCINNATI', u'DENVER', u'DALLAS', u'DETROIT',
          u'FORT WORTH', u'HOUSTON', u'WESTCHESTER COUNTY', u'INDIANAPOLIS', u'NEWARK',
          u'LAS VEGAS', u'LOS ANGELES', u'LONG BEACH', u'ATLANTA', u'MEMPHIS', u'MIAMI',
          u'KANSAS CITY', u'MILWAUKEE', u'MINNEAPOLIS', u'NEW YORK', u'OAKLAND', u'ONTARIO',
          u'ORLANDO', u'PHILADELPHIA', u'PHOENIX', u'PITTSBURGH', u'ST. PAUL', u'SAN DIEGO',
          u'SEATTLE', u'SAN FRANCISCO', u'SAN JOSE', u'SALT LAKE CITY', u'ST. LOUIS',
          u'ST. PETERSBURG', u'TACOMA', u'TAMPA', u'WASHINGTON', u'MONTREAL', u'TORONTO']

CITY_CODE_LIST = [u'BBNA', u'BBOS', u'BBUR', u'BBWI', u'CCHI', u'CCLE', u'CCLT', u'CCMH', u'CCVG', u'DDEN',
                  u'DDFW', u'DDTT', u'FDFW', u'HHOU', u'HHPN', u'IIND', u'JNYC', u'LLAS', u'LLAX', u'LLGB',
                  u'MATL', u'MMEM', u'MMIA', u'MMKC', u'MMKE', u'MMSP', u'NNYC', u'OOAK', u'OONT', u'OORL',
                  u'PPHL', u'PPHX', u'PPIT', u'SMSP', u'SSAN', u'SSEA', u'SSFO', u'SSJC', u'SSLC', u'SSTL',
                  u'STPA', u'TSEA', u'TTPA', u'WWAS', u'YYMQ', u'YYTO']

CLASS = [u'COACH', u'BUSINESS', u'FIRST', u'THRIST', u'STANDARD', u'SHUTTLE']

DAY_OF_WEEK = [u'MONDAY', u'TUESDAY', u'WEDNESDAY', u'THURSDAY', u'FRIDAY', u'SATURDAY', u'SUNDAY']

FARE_BASIS_CODE = [u'B', u'BH', u'BHW', u'BHX', u'BL', u'BLW', u'BLX', u'BN', u'BOW', u'BOX',
                   u'BW', u'BX', u'C', u'CN', u'F', u'FN', u'H', u'HH', u'HHW', u'HHX', u'HL', u'HLW', u'HLX',
                   u'HOW', u'HOX', u'J', u'K', u'KH', u'KL', u'KN', u'LX', u'M', u'MH', u'ML', u'MOW', u'P',
                   u'Q', u'QH', u'QHW', u'QHX', u'QLW', u'QLX', u'QO', u'QOW', u'QOX', u'QW', u'QX', u'S',
                   u'U', u'V', u'VHW', u'VHX', u'VW', u'VX', u'Y', u'YH', u'YL', u'YN', u'YW', u'YX']

MEALS = [u'BREAKFAST', u'LUNCH', u'SNACK', u'DINNER']

RESTRICT_CODES = [u'AP/2', u'AP/6', u'AP/12', u'AP/20', u'AP/21', u'AP/57', u'AP/58', u'AP/60',
                  u'AP/75', u'EX/9', u'EX/13', u'EX/14', u'EX/17', u'EX/19']

STATES = [u'ARIZONA', u'CALIFORNIA', u'COLORADO', u'DISTRICT OF COLUMBIA',
          u'FLORIDA', u'GEORGIA', u'ILLINOIS', u'INDIANA', u'MASSACHUSETTS',
          u'MARYLAND', u'MICHIGAN', u'MINNESOTA', u'MISSOURI', u'NORTH CAROLINA',
          u'NEW JERSEY', u'NEVADA', u'NEW YORK', u'OHIO', u'ONTARIO', u'PENNSYLVANIA',
          u'QUEBEC', u'TENNESSEE', u'TEXAS', u'UTAH', u'WASHINGTON', u'WISCONSIN']

STATE_CODES = [u'TN', u'MA', u'CA', u'MD', u'IL', u'OH', u'NC', u'CO', u'TX', u'MI', u'NY',
               u'IN', u'NJ', u'NV', u'GA', u'FL', u'MO', u'WI', u'MN', u'PA', u'AZ', u'WA',
               u'UT', u'DC', u'PQ', u'ON']

DAY_OF_WEEK_INDEX = dict((idx, [day]) for idx, day in enumerate(DAY_OF_WEEK))

TRIGGER_LISTS = [CITIES, AIRPORT_CODES,
                 STATES, STATE_CODES,
                 FARE_BASIS_CODE, CLASS,
                 AIRLINE_CODE_LIST, DAY_OF_WEEK,
                 CITY_CODE_LIST, MEALS,
                 RESTRICT_CODES]

TRIGGER_DICTS = [CITY_AIRPORT_CODES,
                 AIRLINE_CODES,
                 CITY_CODES,
                 GROUND_SERVICE,
                 DAY_OF_WEEK_DICT,
                 YES_NO,
                 MISC_STR]

ATIS_TRIGGER_DICT = get_trigger_dict(TRIGGER_LISTS, TRIGGER_DICTS)

NUMBER_TRIGGER_DICT                       = get_trigger_dict([], [convert_to_string_list_value_dict(MONTH_NUMBERS),
                                                                  convert_to_string_list_value_dict(DAY_NUMBERS),
                                                                  MISC_TIME_TRIGGERS])




from __future__ import absolute_import
#typing

#overrides
import torch
from torch.nn.modules.linear import Linear

from allennlp.common.checks import check_dimensions_match
from allennlp.data import Vocabulary
from allennlp.modules import Seq2SeqEncoder, TimeDistributed, TextFieldEmbedder
from allennlp.modules import ConditionalRandomField, FeedForward
from allennlp.modules.conditional_random_field import allowed_transitions
from allennlp.models.model import Model
from allennlp.nn import InitializerApplicator, RegularizerApplicator
import allennlp.nn.util as util
from allennlp.training.metrics import SpanBasedF1Measure


class CrfTagger(Model):
    u"""
    The ``CrfTagger`` encodes a sequence of text with a ``Seq2SeqEncoder``,
    then uses a Conditional Random Field model to predict a tag for each token in the sequence.

    Parameters
    ----------
    vocab : ``Vocabulary``, required
        A Vocabulary, required in order to compute sizes for input/output projections.
    text_field_embedder : ``TextFieldEmbedder``, required
        Used to embed the tokens ``TextField`` we get as input to the model.
    encoder : ``Seq2SeqEncoder``
        The encoder that we will use in between embedding tokens and predicting output tags.
    label_namespace : ``str``, optional (default=``labels``)
        This is needed to compute the SpanBasedF1Measure metric.
        Unless you did something unusual, the default value should be what you want.
    feedforward : ``FeedForward``, optional, (default = None).
        An optional feedforward layer to apply after the encoder.
    dropout:  ``float``, optional (detault=``None``)
    verbose_metrics : ``bool``, optional (default = False)
        If true, metrics will be returned per label class in addition
        to the overall statistics.
    constraint_type : ``str``, optional (default=``None``)
        If provided, the CRF will be constrained at decoding time
        to produce valid labels based on the specified type (e.g. "BIO", or "BIOUL").
    include_start_end_transitions : ``bool``, optional (default=``True``)
        Whether to include start and end transition parameters in the CRF.
    initializer : ``InitializerApplicator``, optional (default=``InitializerApplicator()``)
        Used to initialize the model parameters.
    regularizer : ``RegularizerApplicator``, optional (default=``None``)
        If provided, will be used to calculate the regularization penalty during training.
    """

    def __init__(self, vocab            ,
                 text_field_embedder                   ,
                 encoder                ,
                 label_namespace      = u"labels",
                 constraint_type      = None,
                 feedforward              = None,
                 include_start_end_transitions       = True,
                 dropout        = None,
                 verbose_metrics       = False,
                 initializer                        = InitializerApplicator(),
                 regularizer                                  = None)        :
        super(CrfTagger, self).__init__(vocab, regularizer)

        self.label_namespace = label_namespace
        self.text_field_embedder = text_field_embedder
        self.num_tags = self.vocab.get_vocab_size(label_namespace)
        self.encoder = encoder
        self._verbose_metrics = verbose_metrics
        if dropout:
            self.dropout = torch.nn.Dropout(dropout)
        else:
            self.dropout = None
        self._feedforward = feedforward

        if feedforward is not None:
            output_dim = feedforward.get_output_dim()
        else:
            output_dim = self.encoder.get_output_dim()
        self.tag_projection_layer = TimeDistributed(Linear(output_dim,
                                                           self.num_tags))

        if constraint_type is not None:
            labels = self.vocab.get_index_to_token_vocabulary(label_namespace)
            constraints = allowed_transitions(constraint_type, labels)
        else:
            constraints = None

        self.crf = ConditionalRandomField(
                self.num_tags, constraints,
                include_start_end_transitions=include_start_end_transitions
        )

        self.span_metric = SpanBasedF1Measure(vocab,
                                              tag_namespace=label_namespace,
                                              label_encoding=constraint_type or u"BIO")


        check_dimensions_match(text_field_embedder.get_output_dim(), encoder.get_input_dim(),
                               u"text field embedding dim", u"encoder input dim")
        if feedforward is not None:
            check_dimensions_match(encoder.get_output_dim(), feedforward.get_input_dim(),
                                   u"encoder output dim", u"feedforward input dim")
        initializer(self)

    #overrides
    def forward(self,  # type: ignore
                tokens                             ,
                tags                   = None,
                metadata                       = None)                           :
        # pylint: disable=arguments-differ
        u"""
        Parameters
        ----------
        tokens : ``Dict[str, torch.LongTensor]``, required
            The output of ``TextField.as_array()``, which should typically be passed directly to a
            ``TextFieldEmbedder``. This output is a dictionary mapping keys to ``TokenIndexer``
            tensors.  At its most basic, using a ``SingleIdTokenIndexer`` this is: ``{"tokens":
            Tensor(batch_size, num_tokens)}``. This dictionary will have the same keys as were used
            for the ``TokenIndexers`` when you created the ``TextField`` representing your
            sequence.  The dictionary is designed to be passed directly to a ``TextFieldEmbedder``,
            which knows how to combine different word representations into a single vector per
            token in your input.
        tags : ``torch.LongTensor``, optional (default = ``None``)
            A torch tensor representing the sequence of integer gold class labels of shape
            ``(batch_size, num_tokens)``.
        metadata : ``List[Dict[str, Any]]``, optional, (default = None)
            metadata containg the original words in the sentence to be tagged under a 'words' key.

        Returns
        -------
        An output dictionary consisting of:

        logits : ``torch.FloatTensor``
            The logits that are the output of the ``tag_projection_layer``
        mask : ``torch.LongTensor``
            The text field mask for the input tokens
        tags : ``List[List[int]]``
            The predicted tags using the Viterbi algorithm.
        loss : ``torch.FloatTensor``, optional
            A scalar loss to be optimised. Only computed if gold label ``tags`` are provided.
        """
        embedded_text_input = self.text_field_embedder(tokens)
        mask = util.get_text_field_mask(tokens)

        if self.dropout:
            embedded_text_input = self.dropout(embedded_text_input)

        encoded_text = self.encoder(embedded_text_input, mask)

        if self.dropout:
            encoded_text = self.dropout(encoded_text)

        if self._feedforward is not None:
            encoded_text = self._feedforward(encoded_text)

        logits = self.tag_projection_layer(encoded_text)
        best_paths = self.crf.viterbi_tags(logits, mask)

        # Just get the tags and ignore the score.
        predicted_tags = [x for x, y in best_paths]

        output = {u"logits": logits, u"mask": mask, u"tags": predicted_tags}

        if tags is not None:
            # Add negative log-likelihood as loss
            log_likelihood = self.crf(logits, tags, mask)
            output[u"loss"] = -log_likelihood

            # Represent viterbi tags as "class probabilities" that we can
            # feed into the `span_metric`
            class_probabilities = logits * 0.
            for i, instance_tags in enumerate(predicted_tags):
                for j, tag_id in enumerate(instance_tags):
                    class_probabilities[i, j, tag_id] = 1

            self.span_metric(class_probabilities, tags, mask)
        if metadata is not None:
            output[u"words"] = [x[u"words"] for x in metadata]
        return output

    #overrides
    def decode(self, output_dict                         )                           :
        u"""
        Converts the tag ids to the actual tags.
        ``output_dict["tags"]`` is a list of lists of tag_ids,
        so we use an ugly nested list comprehension.
        """
        output_dict[u"tags"] = [
                [self.vocab.get_token_from_index(tag, namespace=self.label_namespace)
                 for tag in instance_tags]
                for instance_tags in output_dict[u"tags"]
        ]

        return output_dict

    #overrides
    def get_metrics(self, reset       = False)                    :
        metric_dict = self.span_metric.get_metric(reset=reset)
        if self._verbose_metrics:
            return metric_dict
        else:
            return dict((x, y) for x, y in list(metric_dict.items()) if u"overall" in x)

CrfTagger = Model.register(u"crf_tagger")(CrfTagger)



u"""
Helper functions for archiving models and restoring archived models.
"""



from __future__ import with_statement
from __future__ import absolute_import
#typing
import json
import logging
import os
import tempfile
import tarfile
import shutil

from allennlp.common.file_utils import cached_path
from allennlp.common.params import Params, unflatten, with_fallback, parse_overrides
from allennlp.models.model import Model, _DEFAULT_WEIGHTS
from io import open

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

# We archive a model by creating a tar.gz file with its weights, config, and vocabulary.
#
# We also may include other arbitrary files in the archive. In this case we store
# the mapping { flattened_path -> filename } in ``files_to_archive.json`` and the files
# themselves under the path ``fta/`` .
#
# These constants are the *known names* under which we archive them.
CONFIG_NAME = u"config.json"
_WEIGHTS_NAME = u"weights.th"
_FTA_NAME = u"files_to_archive.json"

def archive_model(serialization_dir     ,
                  weights      = _DEFAULT_WEIGHTS,
                  files_to_archive                 = None)        :
    u"""
    Archive the model weights, its training configuration, and its
    vocabulary to `model.tar.gz`. Include the additional ``files_to_archive``
    if provided.

    Parameters
    ----------
    serialization_dir: ``str``
        The directory where the weights and vocabulary are written out.
    weights: ``str``, optional (default=_DEFAULT_WEIGHTS)
        Which weights file to include in the archive. The default is ``best.th``.
    files_to_archive: ``Dict[str, str]``, optional (default=None)
        A mapping {flattened_key -> filename} of supplementary files to include
        in the archive. That is, if you wanted to include ``params['model']['weights']``
        then you would specify the key as `"model.weights"`.
    """
    weights_file = os.path.join(serialization_dir, weights)
    if not os.path.exists(weights_file):
        logger.error(u"weights file %s does not exist, unable to archive model", weights_file)
        return

    config_file = os.path.join(serialization_dir, CONFIG_NAME)
    if not os.path.exists(config_file):
        logger.error(u"config file %s does not exist, unable to archive model", config_file)

    # If there are files we want to archive, write out the mapping
    # so that we can use it during de-archiving.
    if files_to_archive:
        fta_filename = os.path.join(serialization_dir, _FTA_NAME)
        with open(fta_filename, u'w') as fta_file:
            fta_file.write(json.dumps(files_to_archive))


    archive_file = os.path.join(serialization_dir, u"model.tar.gz")
    logger.info(u"archiving weights and vocabulary to %s", archive_file)
    with tarfile.open(archive_file, u'w:gz') as archive:
        archive.add(config_file, arcname=CONFIG_NAME)
        archive.add(weights_file, arcname=_WEIGHTS_NAME)
        archive.add(os.path.join(serialization_dir, u"vocabulary"),
                    arcname=u"vocabulary")

        # If there are supplemental files to archive:
        if files_to_archive:
            # Archive the { flattened_key -> original_filename } mapping.
            archive.add(fta_filename, arcname=_FTA_NAME)
            # And add each requested file to the archive.
            for key, filename in list(files_to_archive.items()):
                archive.add(filename, arcname="fta/{key}")

def load_archive(archive_file     ,
                 cuda_device      = -1,
                 overrides      = u"",
                 weights_file      = None)           :
    u"""
    Instantiates an Archive from an archived `tar.gz` file.

    Parameters
    ----------
    archive_file: ``str``
        The archive file to load the model from.
    weights_file: ``str``, optional (default = None)
        The weights file to use.  If unspecified, weights.th in the archive_file will be used.
    cuda_device: ``int``, optional (default = -1)
        If `cuda_device` is >= 0, the model will be loaded onto the
        corresponding GPU. Otherwise it will be loaded onto the CPU.
    overrides: ``str``, optional (default = "")
        JSON overrides to apply to the unarchived ``Params`` object.
    """
    # redirect to the cache, if necessary
    resolved_archive_file = cached_path(archive_file)

    if resolved_archive_file == archive_file:
        logger.info("loading archive file {archive_file}")
    else:
        logger.info("loading archive file {archive_file} from cache at {resolved_archive_file}")

    tempdir = None
    if os.path.isdir(resolved_archive_file):
        serialization_dir = resolved_archive_file
    else:
        # Extract archive to temp dir
        tempdir = tempfile.mkdtemp()
        logger.info("extracting archive file {resolved_archive_file} to temp dir {tempdir}")
        with tarfile.open(resolved_archive_file, u'r:gz') as archive:
            archive.extractall(tempdir)

        serialization_dir = tempdir

    # Check for supplemental files in archive
    fta_filename = os.path.join(serialization_dir, _FTA_NAME)
    if os.path.exists(fta_filename):
        with open(fta_filename, u'r') as fta_file:
            files_to_archive = json.loads(fta_file.read())

        # Add these replacements to overrides
        replacements_dict                 = {}
        for key, _ in list(files_to_archive.items()):
            replacement_filename = os.path.join(serialization_dir, "fta/{key}")
            replacements_dict[key] = replacement_filename

        overrides_dict = parse_overrides(overrides)
        combined_dict = with_fallback(preferred=unflatten(replacements_dict), fallback=overrides_dict)
        overrides = json.dumps(combined_dict)

    # Load config
    config = Params.from_file(os.path.join(serialization_dir, CONFIG_NAME), overrides)
    config.loading_from_archive = True

    if weights_file:
        weights_path = weights_file
    else:
        weights_path = os.path.join(serialization_dir, _WEIGHTS_NAME)

    # Instantiate model. Use a duplicate of the config, as it will get consumed.
    model = Model.load(config.duplicate(),
                       weights_file=weights_path,
                       serialization_dir=serialization_dir,
                       cuda_device=cuda_device)

    if tempdir:
        # Clean up temp dir
        shutil.rmtree(tempdir)

    return Archive(model=model, config=config)



# pylint: disable=no-self-use,invalid-name


from __future__ import with_statement
from __future__ import absolute_import
import pytest
import numpy

from allennlp.common.checks import ConfigurationError
from allennlp.common.testing import AllenNlpTestCase
from allennlp.data import Instance, Token, Vocabulary
from allennlp.data.dataset import Batch
from allennlp.data.fields import TextField, LabelField
from allennlp.data.token_indexers import SingleIdTokenIndexer


class TestDataset(AllenNlpTestCase):
    def setUp(self):
        self.vocab = Vocabulary()
        self.vocab.add_token_to_namespace(u"this")
        self.vocab.add_token_to_namespace(u"is")
        self.vocab.add_token_to_namespace(u"a")
        self.vocab.add_token_to_namespace(u"sentence")
        self.vocab.add_token_to_namespace(u".")
        self.token_indexer = {u"tokens": SingleIdTokenIndexer()}
        self.instances = self.get_instances()
        super(TestDataset, self).setUp()

    def test_instances_must_have_homogeneous_fields(self):
        instance1 = Instance({u"tag": (LabelField(1, skip_indexing=True))})
        instance2 = Instance({u"words": TextField([Token(u"hello")], {})})
        with pytest.raises(ConfigurationError):
            _ = Batch([instance1, instance2])

    def test_padding_lengths_uses_max_instance_lengths(self):
        dataset = Batch(self.instances)
        dataset.index_instances(self.vocab)
        padding_lengths = dataset.get_padding_lengths()
        assert padding_lengths == {u"text1": {u"num_tokens": 5}, u"text2": {u"num_tokens": 6}}

    def test_as_tensor_dict(self):
        dataset = Batch(self.instances)
        dataset.index_instances(self.vocab)
        padding_lengths = dataset.get_padding_lengths()
        tensors = dataset.as_tensor_dict(padding_lengths)
        text1 = tensors[u"text1"][u"tokens"].detach().cpu().numpy()
        text2 = tensors[u"text2"][u"tokens"].detach().cpu().numpy()

        numpy.testing.assert_array_almost_equal(text1, numpy.array([[2, 3, 4, 5, 6],
                                                                    [1, 3, 4, 5, 6]]))
        numpy.testing.assert_array_almost_equal(text2, numpy.array([[2, 3, 4, 1, 5, 6],
                                                                    [2, 3, 1, 0, 0, 0]]))

    def get_instances(self):
        field1 = TextField([Token(t) for t in [u"this", u"is", u"a", u"sentence", u"."]],
                           self.token_indexer)
        field2 = TextField([Token(t) for t in [u"this", u"is", u"a", u"different", u"sentence", u"."]],
                           self.token_indexer)
        field3 = TextField([Token(t) for t in [u"here", u"is", u"a", u"sentence", u"."]],
                           self.token_indexer)
        field4 = TextField([Token(t) for t in [u"this", u"is", u"short"]],
                           self.token_indexer)
        instances = [Instance({u"text1": field1, u"text2": field2}),
                     Instance({u"text1": field3, u"text2": field4})]
        return instances



u"""
These submodules contain the classes for AllenNLP models,
all of which are subclasses of :class:`~allennlp.models.model.Model`.
"""


from __future__ import absolute_import
from allennlp.models.model import Model
from allennlp.models.archival import archive_model, load_archive
# from allennlp.models.biattentive_classification_network import BiattentiveClassificationNetwork
# from allennlp.models.constituency_parser import SpanConstituencyParser
# from allennlp.models.biaffine_dependency_parser import BiaffineDependencyParser
# from allennlp.models.coreference_resolution.coref import CoreferenceResolver
# from allennlp.models.crf_tagger import CrfTagger
# from allennlp.models.decomposable_attention import DecomposableAttention
# from allennlp.models.encoder_decoders.simple_seq2seq import SimpleSeq2Seq
# from allennlp.models.reading_comprehension.bidaf import BidirectionalAttentionFlow
# from allennlp.models.semantic_parsing.nlvr.nlvr_coverage_semantic_parser import NlvrCoverageSemanticParser
# from allennlp.models.semantic_parsing.nlvr.nlvr_direct_semantic_parser import NlvrDirectSemanticParser
# from allennlp.models.semantic_parsing.wikitables.wikitables_mml_semantic_parser import WikiTablesMmlSemanticParser
# from allennlp.models.semantic_parsing.wikitables.wikitables_erm_semantic_parser import WikiTablesErmSemanticParser
# from allennlp.models.semantic_role_labeler import SemanticRoleLabeler
# from allennlp.models.simple_tagger import SimpleTagger
# from allennlp.models.esim import ESIM
# from allennlp.models.bimpm import BiMpm



u"""
This module contains various classes for performing
tokenization, stemming, and filtering.
"""


from __future__ import absolute_import
from allennlp.data.tokenizers.tokenizer import Token, Tokenizer
from allennlp.data.tokenizers.word_tokenizer import WordTokenizer
from allennlp.data.tokenizers.character_tokenizer import CharacterTokenizer



# pylint: disable=invalid-name



from __future__ import with_statement
from __future__ import division
from __future__ import absolute_import
import glob
import os
import re
import time

import torch
import pytest
from allennlp.common.checks import ConfigurationError

from allennlp.common.testing import AllenNlpTestCase
from allennlp.training.trainer import Trainer, sparse_clip_norm, is_sparse
from allennlp.training.learning_rate_schedulers import LearningRateScheduler
from allennlp.data import Vocabulary
from allennlp.common.params import Params
from allennlp.models.simple_tagger import SimpleTagger
from allennlp.data.iterators import BasicIterator
from allennlp.data.dataset_readers import SequenceTaggingDatasetReader


class TestTrainer(AllenNlpTestCase):
    def setUp(self):
        super(TestTrainer, self).setUp()
        self.instances = SequenceTaggingDatasetReader().read(self.FIXTURES_ROOT / u'data' / u'sequence_tagging.tsv')
        vocab = Vocabulary.from_instances(self.instances)
        self.vocab = vocab
        self.model_params = Params({
                u"text_field_embedder": {
                        u"tokens": {
                                u"type": u"embedding",
                                u"embedding_dim": 5
                                }
                        },
                u"encoder": {
                        u"type": u"lstm",
                        u"input_size": 5,
                        u"hidden_size": 7,
                        u"num_layers": 2
                        }
                })
        self.model = SimpleTagger.from_params(vocab=self.vocab, params=self.model_params)
        self.optimizer = torch.optim.SGD(self.model.parameters(), 0.01)
        self.iterator = BasicIterator(batch_size=2)
        self.iterator.index_with(vocab)

    def test_trainer_can_run(self):
        trainer = Trainer(model=self.model,
                          optimizer=self.optimizer,
                          iterator=self.iterator,
                          train_dataset=self.instances,
                          validation_dataset=self.instances,
                          num_epochs=2)
        metrics = trainer.train()
        assert u'best_validation_loss' in metrics
        assert isinstance(metrics[u'best_validation_loss'], float)
        assert u'best_validation_accuracy' in metrics
        assert isinstance(metrics[u'best_validation_accuracy'], float)
        assert u'best_validation_accuracy3' in metrics
        assert isinstance(metrics[u'best_validation_accuracy3'], float)
        assert u'best_epoch' in metrics
        assert isinstance(metrics[u'best_epoch'], int)

        # Making sure that both increasing and decreasing validation metrics work.
        trainer = Trainer(model=self.model,
                          optimizer=self.optimizer,
                          iterator=self.iterator,
                          train_dataset=self.instances,
                          validation_dataset=self.instances,
                          validation_metric=u'+loss',
                          num_epochs=2)
        metrics = trainer.train()
        assert u'best_validation_loss' in metrics
        assert isinstance(metrics[u'best_validation_loss'], float)
        assert u'best_validation_accuracy' in metrics
        assert isinstance(metrics[u'best_validation_accuracy'], float)
        assert u'best_validation_accuracy3' in metrics
        assert isinstance(metrics[u'best_validation_accuracy3'], float)
        assert u'best_epoch' in metrics
        assert isinstance(metrics[u'best_epoch'], int)

    @pytest.mark.skipif(not torch.cuda.is_available(), reason=u"No CUDA device registered.")
    def test_trainer_can_run_cuda(self):
        trainer = Trainer(self.model, self.optimizer,
                          self.iterator, self.instances, num_epochs=2,
                          cuda_device=0)
        trainer.train()

    @pytest.mark.skipif(torch.cuda.device_count() < 2,
                        reason=u"Need multiple GPUs.")
    def test_trainer_can_run_multiple_gpu(self):
        multigpu_iterator = BasicIterator(batch_size=4)
        multigpu_iterator.index_with(self.vocab)
        trainer = Trainer(self.model, self.optimizer,
                          multigpu_iterator, self.instances, num_epochs=2,
                          cuda_device=[0, 1])
        trainer.train()

    def test_trainer_can_resume_training(self):
        trainer = Trainer(self.model, self.optimizer,
                          self.iterator, self.instances,
                          validation_dataset=self.instances,
                          num_epochs=1, serialization_dir=self.TEST_DIR)
        trainer.train()
        new_trainer = Trainer(self.model, self.optimizer,
                              self.iterator, self.instances,
                              validation_dataset=self.instances,
                              num_epochs=3, serialization_dir=self.TEST_DIR)

        epoch, val_metrics_per_epoch = new_trainer._restore_checkpoint()  # pylint: disable=protected-access
        assert epoch == 1
        assert len(val_metrics_per_epoch) == 1
        assert isinstance(val_metrics_per_epoch[0], float)
        assert val_metrics_per_epoch[0] != 0.
        new_trainer.train()

    def test_metric_only_considered_best_so_far_when_strictly_better_than_those_before_it_increasing_metric(
            self):
        new_trainer = Trainer(self.model, self.optimizer,
                              self.iterator, self.instances,
                              validation_dataset=self.instances,
                              num_epochs=3, serialization_dir=self.TEST_DIR,
                              patience=5, validation_metric=u"+test")
        # when it is the only metric it should be considered the best
        assert new_trainer._is_best_so_far(1, [])  # pylint: disable=protected-access
        # when it is the same as one before it it is not considered the best
        assert not new_trainer._is_best_so_far(.3, [.3, .3, .3, .2, .5, .1])  # pylint: disable=protected-access
        # when it is the best it is considered the best
        assert new_trainer._is_best_so_far(13.00, [.3, .3, .3, .2, .5, .1])  # pylint: disable=protected-access
        # when it is not the the best it is not considered the best
        assert not new_trainer._is_best_so_far(.0013, [.3, .3, .3, .2, .5, .1])  # pylint: disable=protected-access

    def test_metric_only_considered_best_so_far_when_strictly_better_than_those_before_it_decreasing_metric(self):
        new_trainer = Trainer(self.model, self.optimizer,
                              self.iterator, self.instances,
                              validation_dataset=self.instances,
                              num_epochs=3, serialization_dir=self.TEST_DIR,
                              patience=5, validation_metric=u"-test")
        # when it is the only metric it should be considered the best
        assert new_trainer._is_best_so_far(1, [])  # pylint: disable=protected-access
        # when it is the same as one before it it is not considered the best
        assert not new_trainer._is_best_so_far(.3, [.3, .3, .3, .2, .5, .1])  # pylint: disable=protected-access
        # when it is the best it is considered the best
        assert new_trainer._is_best_so_far(.013, [.3, .3, .3, .2, .5, .1])  # pylint: disable=protected-access
        # when it is not the the best it is not considered the best
        assert not new_trainer._is_best_so_far(13.00, [.3, .3, .3, .2, .5, .1])  # pylint: disable=protected-access

    def test_should_stop_early_with_increasing_metric(self):
        new_trainer = Trainer(self.model, self.optimizer,
                              self.iterator, self.instances,
                              validation_dataset=self.instances,
                              num_epochs=3, serialization_dir=self.TEST_DIR,
                              patience=5, validation_metric=u"+test")
        assert new_trainer._should_stop_early([.5, .3, .2, .1, .4, .4])  # pylint: disable=protected-access
        assert not new_trainer._should_stop_early([.3, .3, .3, .2, .5, .1])  # pylint: disable=protected-access

    def test_should_stop_early_with_flat_lining_metric(self):
        flatline = [.2] * 6
        assert Trainer(self.model, self.optimizer,  # pylint: disable=protected-access
                       self.iterator, self.instances,
                       validation_dataset=self.instances,
                       num_epochs=3,
                       serialization_dir=self.TEST_DIR,
                       patience=5,
                       validation_metric=u"+test")._should_stop_early(flatline)  # pylint: disable=protected-access
        assert Trainer(self.model, self.optimizer,  # pylint: disable=protected-access
                       self.iterator, self.instances,
                       validation_dataset=self.instances,
                       num_epochs=3,
                       serialization_dir=self.TEST_DIR,
                       patience=5,
                       validation_metric=u"-test")._should_stop_early(flatline)  # pylint: disable=protected-access

    def test_should_stop_early_with_decreasing_metric(self):
        new_trainer = Trainer(self.model, self.optimizer,
                              self.iterator, self.instances,
                              validation_dataset=self.instances,
                              num_epochs=3, serialization_dir=self.TEST_DIR,
                              patience=5, validation_metric=u"-test")
        assert new_trainer._should_stop_early([.02, .3, .2, .1, .4, .4])  # pylint: disable=protected-access
        assert not new_trainer._should_stop_early([.3, .3, .2, .1, .4, .5])  # pylint: disable=protected-access
        assert new_trainer._should_stop_early([.1, .3, .2, .1, .4, .5])  # pylint: disable=protected-access

    def test_should_stop_early_with_early_stopping_disabled(self):
        # Increasing metric
        trainer = Trainer(self.model, self.optimizer, self.iterator, self.instances,
                          validation_dataset=self.instances, num_epochs=100,
                          patience=None, validation_metric=u"+test")
        decreasing_history = [float(i) for i in reversed(range(20))]
        assert not trainer._should_stop_early(decreasing_history)  # pylint: disable=protected-access

        # Decreasing metric
        trainer = Trainer(self.model, self.optimizer, self.iterator, self.instances,
                          validation_dataset=self.instances, num_epochs=100,
                          patience=None, validation_metric=u"-test")
        increasing_history = [float(i) for i in range(20)]
        assert not trainer._should_stop_early(increasing_history)  # pylint: disable=protected-access

    def test_should_stop_early_with_invalid_patience(self):
        for patience in [0, -1, -2, 1.5, u'None']:
            with pytest.raises(ConfigurationError,
                               message=u'No ConfigurationError for patience={}'.format(patience)):
                Trainer(self.model, self.optimizer, self.iterator, self.instances,
                        validation_dataset=self.instances, num_epochs=100,
                        patience=patience, validation_metric=u"+test")

    def test_trainer_can_run_with_lr_scheduler(self):

        lr_params = Params({u"type": u"reduce_on_plateau"})
        lr_scheduler = LearningRateScheduler.from_params(self.optimizer, lr_params)
        trainer = Trainer(model=self.model,
                          optimizer=self.optimizer,
                          iterator=self.iterator,
                          learning_rate_scheduler=lr_scheduler,
                          validation_metric=u"-loss",
                          train_dataset=self.instances,
                          validation_dataset=self.instances,
                          num_epochs=2)
        trainer.train()

    def test_trainer_raises_on_model_with_no_loss_key(self):
        class FakeModel(torch.nn.Module):
            def forward(self, **kwargs):  # pylint: disable=arguments-differ,unused-argument
                return {}
        with pytest.raises(RuntimeError):
            trainer = Trainer(FakeModel(), self.optimizer,
                              self.iterator, self.instances,
                              num_epochs=2, serialization_dir=self.TEST_DIR)
            trainer.train()

    def test_trainer_can_log_histograms(self):
        # enable activation logging
        for module in self.model.modules():
            module.should_log_activations = True

        trainer = Trainer(self.model, self.optimizer,
                          self.iterator, self.instances, num_epochs=3,
                          serialization_dir=self.TEST_DIR,
                          histogram_interval=2)
        trainer.train()

    def test_trainer_respects_num_serialized_models_to_keep(self):
        trainer = Trainer(self.model, self.optimizer,
                          self.iterator, self.instances, num_epochs=5,
                          serialization_dir=self.TEST_DIR,
                          num_serialized_models_to_keep=3)
        trainer.train()

        # Now check the serialized files
        for prefix in [u'model_state_epoch_*', u'training_state_epoch_*']:
            file_names = glob.glob(os.path.join(self.TEST_DIR, prefix))
            epochs = [int(re.search(ur"_([0-9])\.th", fname).group(1))
                      for fname in file_names]
            assert sorted(epochs) == [2, 3, 4]

    def test_trainer_respects_keep_serialized_model_every_num_seconds(self):
        # To test:
        #   Create an iterator that sleeps for 2.5 second per epoch, so the total training
        #       time for one epoch is slightly greater then 2.5 seconds.
        #   Run for 6 epochs, keeping the last 2 models, models also kept every 5 seconds.
        #   Check the resulting checkpoints.  Should then have models at epochs
        #       2, 4, plus the last two at 5 and 6.
        class WaitingIterator(BasicIterator):
            # pylint: disable=arguments-differ
            def _create_batches(self, *args, **kwargs):
                time.sleep(2.5)
                return super(WaitingIterator, self)._create_batches(*args, **kwargs)

        iterator = WaitingIterator(batch_size=2)
        iterator.index_with(self.vocab)

        trainer = Trainer(self.model, self.optimizer,
                          iterator, self.instances, num_epochs=6,
                          serialization_dir=self.TEST_DIR,
                          num_serialized_models_to_keep=2,
                          keep_serialized_model_every_num_seconds=5)
        trainer.train()

        # Now check the serialized files
        for prefix in [u'model_state_epoch_*', u'training_state_epoch_*']:
            file_names = glob.glob(os.path.join(self.TEST_DIR, prefix))
            epochs = [int(re.search(ur"_([0-9])\.th", fname).group(1))
                      for fname in file_names]
            # epoch N has N-1 in file name
            assert sorted(epochs) == [1, 3, 4, 5]

    def test_trainer_saves_models_at_specified_interval(self):
        iterator = BasicIterator(batch_size=4)
        iterator.index_with(self.vocab)

        trainer = Trainer(self.model, self.optimizer,
                          iterator, self.instances, num_epochs=2,
                          serialization_dir=self.TEST_DIR,
                          model_save_interval=0.0001)

        trainer.train()

        # Now check the serialized files for models saved during the epoch.
        prefix = u'model_state_epoch_*'
        file_names = sorted(glob.glob(os.path.join(self.TEST_DIR, prefix)))
        epochs = [re.search(ur"_([0-9\.\-]+)\.th", fname).group(1)
                  for fname in file_names]
        # We should have checkpoints at the end of each epoch and during each, e.g.
        # [0.timestamp, 0, 1.timestamp, 1]
        assert len(epochs) == 4
        assert epochs[3] == u'1'
        assert u'.' in epochs[0]

        # Now make certain we can restore from timestamped checkpoint.
        # To do so, remove the checkpoint from the end of epoch 1&2, so
        # that we are forced to restore from the timestamped checkpoints.
        for k in range(2):
            os.remove(os.path.join(self.TEST_DIR, u'model_state_epoch_{}.th'.format(k)))
            os.remove(os.path.join(self.TEST_DIR, u'training_state_epoch_{}.th'.format(k)))
        os.remove(os.path.join(self.TEST_DIR, u'best.th'))

        restore_trainer = Trainer(self.model, self.optimizer,
                                  self.iterator, self.instances, num_epochs=2,
                                  serialization_dir=self.TEST_DIR,
                                  model_save_interval=0.0001)
        epoch, _ = restore_trainer._restore_checkpoint()  # pylint: disable=protected-access
        assert epoch == 2
        # One batch per epoch.
        assert restore_trainer._batch_num_total == 2  # pylint: disable=protected-access


class TestSparseClipGrad(AllenNlpTestCase):
    def test_sparse_clip_grad(self):
        # create a sparse embedding layer, then take gradient
        embedding = torch.nn.Embedding(100, 16, sparse=True)
        embedding.zero_grad()
        ids = (torch.rand(17) * 100).long()
        # Set some of the ids to the same value so that the sparse gradient
        # has repeated indices.  This tests some additional logic.
        ids[:5] = 5
        loss = embedding(ids).sum()
        loss.backward()
        assert is_sparse(embedding.weight.grad)

        # Now try to clip the gradients.
        _ = sparse_clip_norm([embedding.weight], 1.5)
        # Final norm should be 1.5
        grad = embedding.weight.grad.coalesce()  # pylint: disable=no-member
        self.assertAlmostEqual(grad._values().norm(2.0).item(), 1.5, places=5) # pylint: disable=protected-access




from __future__ import absolute_import
#typing
import logging
import random

from allennlp.common.util import lazy_groups_of
from allennlp.data.instance import Instance
from allennlp.data.iterators.data_iterator import DataIterator
from allennlp.data.dataset import Batch

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class BasicIterator(DataIterator):
    u"""
    A very basic iterator that takes a dataset, possibly shuffles it, and creates fixed sized batches.

    It takes the same parameters as :class:`allennlp.data.iterators.DataIterator`
    """
    def _create_batches(self, instances                    , shuffle      )                   :
        # First break the dataset into memory-sized lists:
        for instance_list in self._memory_sized_lists(instances):
            if shuffle:
                random.shuffle(instance_list)
            iterator = iter(instance_list)
            # Then break each memory-sized list into batches.
            for batch_instances in lazy_groups_of(iterator, self._batch_size):
                for possibly_smaller_batches in self._ensure_batch_is_sufficiently_small(batch_instances):
                    batch = Batch(possibly_smaller_batches)
                    yield batch

BasicIterator = DataIterator.register(u"basic")(BasicIterator)



# -*- coding: utf-8 -*-
# pylint: disable=no-self-use,invalid-name



from __future__ import division
from __future__ import with_statement
from __future__ import absolute_import
import json
import os
import pathlib
import sys
import tempfile
import warnings
from io import open
try:
    from itertools import izip
except:
    izip = zip


with warnings.catch_warnings():
    warnings.filterwarnings(u"ignore", category=FutureWarning)
    import h5py
import numpy
import pytest

from allennlp.common.checks import ConfigurationError
from allennlp.commands import main
from allennlp.commands.elmo import ElmoEmbedder
from allennlp.tests.modules.elmo_test import ElmoTestCase


class TestElmoCommand(ElmoTestCase):
    def setUp(self):
        super(TestElmoCommand, self).setUp()
        self.tempdir = pathlib.Path(tempfile.mkdtemp())
        self.sentences_path = unicode(self.tempdir / u"sentences.txt")
        self.output_path = unicode(self.tempdir / u"output.txt")

    def test_all_embedding_works(self):
        sentence = u"Michael went to the store to buy some eggs ."
        with open(self.sentences_path, u'w') as f:
            f.write(sentence)

        sys.argv = [u"run.py",  # executable
                    u"elmo",  # command
                    self.sentences_path,
                    self.output_path,
                    u"--all",
                    u"--options-file",
                    self.options_file,
                    u"--weight-file",
                    self.weight_file]

        main()

        assert os.path.exists(self.output_path)

        embedder = ElmoEmbedder(options_file=self.options_file, weight_file=self.weight_file)
        expected_embedding = embedder.embed_sentence(sentence.split())

        with h5py.File(self.output_path, u'r') as h5py_file:
            assert set(h5py_file.keys()) == set([u"0", u"sentence_to_index"])
            # The vectors in the test configuration are smaller (32 length)
            embedding = h5py_file.get(u"0")
            assert embedding.shape == (3, len(sentence.split()), 32)
            numpy.testing.assert_allclose(embedding, expected_embedding, rtol=1e-4)
            assert json.loads(h5py_file.get(u"sentence_to_index")[0]) == {sentence: u"0"}

    def test_top_embedding_works(self):
        sentence = u"Michael went to the store to buy some eggs ."
        with open(self.sentences_path, u'w') as f:
            f.write(sentence)

        sys.argv = [u"run.py",  # executable
                    u"elmo",  # command
                    self.sentences_path,
                    self.output_path,
                    u"--top",
                    u"--options-file",
                    self.options_file,
                    u"--weight-file",
                    self.weight_file]

        main()

        assert os.path.exists(self.output_path)

        embedder = ElmoEmbedder(options_file=self.options_file, weight_file=self.weight_file)
        expected_embedding = embedder.embed_sentence(sentence.split())[2]

        with h5py.File(self.output_path, u'r') as h5py_file:
            assert set(h5py_file.keys()) == set([u"0", u"sentence_to_index"])
            # The vectors in the test configuration are smaller (32 length)
            embedding = h5py_file.get(u"0")
            assert embedding.shape == (len(sentence.split()), 32)
            numpy.testing.assert_allclose(embedding, expected_embedding, rtol=1e-4)
            assert json.loads(h5py_file.get(u"sentence_to_index")[0]) == {sentence: u"0"}

    def test_average_embedding_works(self):
        sentence = u"Michael went to the store to buy some eggs ."
        with open(self.sentences_path, u'w') as f:
            f.write(sentence)

        sys.argv = [u"run.py",  # executable
                    u"elmo",  # command
                    self.sentences_path,
                    self.output_path,
                    u"--average",
                    u"--options-file",
                    self.options_file,
                    u"--weight-file",
                    self.weight_file]

        main()

        assert os.path.exists(self.output_path)

        embedder = ElmoEmbedder(options_file=self.options_file, weight_file=self.weight_file)
        expected_embedding = embedder.embed_sentence(sentence.split())
        expected_embedding = (expected_embedding[0] + expected_embedding[1] + expected_embedding[2]) / 3

        with h5py.File(self.output_path, u'r') as h5py_file:
            assert set(h5py_file.keys()) == set([u"0", u"sentence_to_index"])
            # The vectors in the test configuration are smaller (32 length)
            embedding = h5py_file.get(u"0")
            assert embedding.shape == (len(sentence.split()), 32)
            numpy.testing.assert_allclose(embedding, expected_embedding, rtol=1e-4)
            assert json.loads(h5py_file.get(u"sentence_to_index")[0]) == {sentence: u"0"}

    def test_batch_embedding_works(self):
        sentences = [
                u"Michael went to the store to buy some eggs .",
                u"Joel rolled down the street on his skateboard .",
                u"test / this is a first sentence",
                u"Take a look , then , at Tuesday 's elections in New York City , New Jersey and Virginia :"
        ]

        with open(self.sentences_path, u'w') as f:
            for line in sentences:
                f.write(line + u'\n')

        sys.argv = [u"run.py",  # executable
                    u"elmo",  # command
                    self.sentences_path,
                    self.output_path,
                    u"--all",
                    u"--options-file",
                    self.options_file,
                    u"--weight-file",
                    self.weight_file]

        main()

        assert os.path.exists(self.output_path)

        with h5py.File(self.output_path, u'r') as h5py_file:
            assert set(h5py_file.keys()) == set([u"0", u"1", u"2", u"3", u"sentence_to_index"])
            # The vectors in the test configuration are smaller (32 length)
            for sentence_id, sentence in izip([u"0", u"1", u"2", u"3"], sentences):
                assert h5py_file.get(sentence_id).shape == (3, len(sentence.split()), 32)
            assert (json.loads(h5py_file.get(u"sentence_to_index")[0]) ==
                    dict((sentences[i], unicode(i)) for i in range(len(sentences))))

    def test_batch_embedding_works_with_sentences_as_keys(self):
        sentences = [
                u"Michael went to the store to buy some eggs .",
                u"Joel rolled down the street on his skateboard ."
        ]

        with open(self.sentences_path, u'w') as f:
            for line in sentences:
                f.write(line + u'\n')

        sys.argv = [u"run.py",  # executable
                    u"elmo",  # command
                    self.sentences_path,
                    self.output_path,
                    u"--all",
                    u"--options-file",
                    self.options_file,
                    u"--weight-file",
                    self.weight_file,
                    u"--use-sentence-keys"]
        main()

        assert os.path.exists(self.output_path)

        with h5py.File(self.output_path, u'r') as h5py_file:
            assert set(h5py_file.keys()) == set(sentences)
            # The vectors in the test configuration are smaller (32 length)
            for sentence in sentences:
                assert h5py_file.get(sentence).shape == (3, len(sentence.split()), 32)

    def test_batch_embedding_works_with_forget_sentences(self):
        sentences = [
                u"Michael went to the store to buy some eggs .",
                u"Joel rolled down the street on his skateboard .",
                u"test / this is a first sentence",
                u"Take a look , then , at Tuesday 's elections in New York City , New Jersey and Virginia :"
        ]

        with open(self.sentences_path, u'w') as f:
            for line in sentences:
                f.write(line + u'\n')

        sys.argv = [u"run.py",  # executable
                    u"elmo",  # command
                    self.sentences_path,
                    self.output_path,
                    u"--all",
                    u"--options-file",
                    self.options_file,
                    u"--weight-file",
                    self.weight_file,
                    u"--forget-sentences"]

        main()

        assert os.path.exists(self.output_path)

        with h5py.File(self.output_path, u'r') as h5py_file:
            assert set(h5py_file.keys()) == set([u"0", u"1", u"2", u"3"])
            # The vectors in the test configuration are smaller (32 length)
            for sentence_id, sentence in izip([u"0", u"1", u"2", u"3"], sentences):
                assert h5py_file.get(sentence_id).shape == (3, len(sentence.split()), 32)

    def test_duplicate_sentences(self):
        sentences = [
                u"Michael went to the store to buy some eggs .",
                u"Michael went to the store to buy some eggs .",
        ]

        with open(self.sentences_path, u'w') as f:
            for line in sentences:
                f.write(line + u'\n')

        sys.argv = [u"run.py",  # executable
                    u"elmo",  # command
                    self.sentences_path,
                    self.output_path,
                    u"--all",
                    u"--options-file",
                    self.options_file,
                    u"--weight-file",
                    self.weight_file]

        main()

        assert os.path.exists(self.output_path)

        with h5py.File(self.output_path, u'r') as h5py_file:
            assert len(list(h5py_file.keys())) == 3
            assert set(h5py_file.keys()) == set([u"0", u"1", u"sentence_to_index"])
            # The vectors in the test configuration are smaller (32 length)
            for sentence_id, sentence in izip([u"0", u"1"], sentences):
                assert h5py_file.get(sentence_id).shape == (3, len(sentence.split()), 32)

    def test_empty_sentences_raise_errors(self):
        sentences = [
                u"A",
                u"",
                u"",
                u"B"
        ]

        with open(self.sentences_path, u'w') as f:
            for line in sentences:
                f.write(line + u'\n')

        sys.argv = [u"run.py",  # executable
                    u"elmo",  # command
                    self.sentences_path,
                    self.output_path,
                    u"--all",
                    u"--options-file",
                    self.options_file,
                    u"--weight-file",
                    self.weight_file]
        with pytest.raises(ConfigurationError):
            main()


class TestElmoEmbedder(ElmoTestCase):
    def test_embeddings_are_as_expected(self):
        loaded_sentences, loaded_embeddings = self._load_sentences_embeddings()

        assert len(loaded_sentences) == len(loaded_embeddings)
        batch_size = len(loaded_sentences)

        # The sentences and embeddings are organized in an idiosyncratic way TensorFlow handles batching.
        # We are going to reorganize them linearly so they can be grouped into batches by AllenNLP.
        sentences = []
        expected_embeddings = []
        for batch_number in range(len(loaded_sentences[0])):
            for index in range(batch_size):
                sentences.append(loaded_sentences[index][batch_number].split())
                expected_embeddings.append(loaded_embeddings[index][batch_number])

        assert len(expected_embeddings) == len(sentences)

        embedder = ElmoEmbedder(options_file=self.options_file, weight_file=self.weight_file)
        embeddings = list(embedder.embed_sentences(sentences, batch_size))

        assert len(embeddings) == len(sentences)

        for tensor, expected in izip(embeddings, expected_embeddings):
            numpy.testing.assert_array_almost_equal(tensor[2], expected)

    def test_embed_batch_is_empty_sentence(self):
        embedder = ElmoEmbedder(options_file=self.options_file, weight_file=self.weight_file)
        embeddings = embedder.embed_sentence([])

        assert embeddings.shape == (3, 0, 1024)

    def test_embed_batch_contains_empty_sentence(self):
        embedder = ElmoEmbedder(options_file=self.options_file, weight_file=self.weight_file)
        embeddings = list(embedder.embed_sentences([u"This is a test".split(), []]))

        assert len(embeddings) == 2



# pylint: disable=no-self-use,invalid-name


from __future__ import division
from __future__ import absolute_import
#typing

import pytest

from allennlp.data.dataset_readers import ConllCorefReader
from allennlp.common.util import ensure_list
from allennlp.common.testing import AllenNlpTestCase
try:
    from itertools import izip
except:
    izip = zip


class TestCorefReader(object):
    span_width = 5

    @pytest.mark.parametrize(u"lazy", (True, False))
    def test_read_from_file(self, lazy):
        conll_reader = ConllCorefReader(max_span_width=self.span_width, lazy=lazy)
        instances = ensure_list(conll_reader.read(unicode(AllenNlpTestCase.FIXTURES_ROOT /
                                                      u'coref' / u'coref.gold_conll')))

        assert len(instances) == 2

        fields = instances[0].fields
        text = [x.text for x in fields[u"text"].tokens]

        assert text == [u'In', u'the', u'summer', u'of', u'2005', u',', u'a', u'picture', u'that',
                        u'people', u'have', u'long', u'been', u'looking', u'forward', u'to',
                        u'started', u'emerging', u'with', u'frequency', u'in', u'various', u'major',
                        u'Hong', u'Kong', u'media', u'.', u'With', u'their', u'unique', u'charm', u',',
                        u'these', u'well', u'-', u'known', u'cartoon', u'images', u'once', u'again',
                        u'caused', u'Hong', u'Kong', u'to', u'be', u'a', u'focus', u'of', u'worldwide',
                        u'attention', u'.', u'The', u'world', u"'s", u'fifth', u'Disney', u'park',
                        u'will', u'soon', u'open', u'to', u'the', u'public', u'here', u'.']

        spans = fields[u"spans"].field_list
        span_starts, span_ends = izip(*[(field.span_start, field.span_end) for field in spans])

        candidate_mentions = self.check_candidate_mentions_are_well_defined(span_starts, span_ends, text)

        gold_span_labels = fields[u"span_labels"]
        gold_indices_with_ids = [(i, x) for i, x in enumerate(gold_span_labels.labels) if x != -1]
        gold_mentions_with_ids                              = [(candidate_mentions[i], x)
                                                               for i, x in gold_indices_with_ids]

        assert ([u"Hong", u"Kong"], 0) in gold_mentions_with_ids
        gold_mentions_with_ids.remove(([u"Hong", u"Kong"], 0))
        assert ([u"Hong", u"Kong"], 0) in gold_mentions_with_ids
        assert ([u"their"], 1) in gold_mentions_with_ids
        # This is a span which exceeds our max_span_width, so it should not be considered.
        assert not ([u"these", u"well", u"known", u"cartoon", u"images"], 1) in gold_mentions_with_ids

        fields = instances[1].fields
        text = [x.text for x in fields[u"text"].tokens]
        assert text == [u'The', u'area', u'of', u'Hong', u'Kong', u'is', u'only', u'one', u'thousand', u'-', u'plus',
                        u'square', u'kilometers', u'.', u'The', u'population', u'is', u'dense', u'.', u'Natural',
                        u'resources', u'are', u'relatively', u'scarce', u'.', u'However', u',', u'the', u'clever',
                        u'Hong', u'Kong', u'people', u'will', u'utilize', u'all', u'resources', u'they', u'have',
                        u'created', u'for', u'developing', u'the', u'Hong', u'Kong', u'tourism', u'industry', u'.']

        spans = fields[u"spans"].field_list
        span_starts, span_ends = izip(*[(field.span_start, field.span_end) for field in spans])

        candidate_mentions = self.check_candidate_mentions_are_well_defined(span_starts, span_ends, text)

        gold_span_labels = fields[u"span_labels"]
        gold_indices_with_ids = [(i, x) for i, x in enumerate(gold_span_labels.labels) if x != -1]
        gold_mentions_with_ids                              = [(candidate_mentions[i], x)
                                                               for i, x in gold_indices_with_ids]

        assert ([u"Hong", u"Kong"], 0) in gold_mentions_with_ids
        gold_mentions_with_ids.remove(([u"Hong", u"Kong"], 0))
        assert ([u"Hong", u"Kong"], 0) in gold_mentions_with_ids
        assert ([u"they"], 1) in gold_mentions_with_ids
        assert ([u'the', u'clever', u'Hong', u'Kong', u'people'], 1) in gold_mentions_with_ids

    def check_candidate_mentions_are_well_defined(self, span_starts, span_ends, text):
        candidate_mentions = []
        for start, end in izip(span_starts, span_ends):
            # Spans are inclusive.
            text_span = text[start: end + 1]
            candidate_mentions.append(text_span)

        # Check we aren't considering zero length spans and all
        # candidate spans are less than what we specified
        assert all([self.span_width >= len(x) > 0 for x in candidate_mentions])  # pylint: disable=len-as-condition
        return candidate_mentions




from __future__ import absolute_import
from allennlp.data.fields.field import Field


class SequenceField(Field):
    u"""
    A ``SequenceField`` represents a sequence of things.  This class just adds a method onto
    ``Field``: :func:`sequence_length`.  It exists so that ``SequenceLabelField``, ``IndexField`` and other
    similar ``Fields`` can have a single type to require, with a consistent API, whether they are
    pointing to words in a ``TextField``, items in a ``ListField``, or something else.
    """
    def sequence_length(self)       :
        u"""
        How many elements are there in this sequence?
        """
        raise NotImplementedError

    def empty_field(self)                   :
        raise NotImplementedError




from __future__ import absolute_import
#overrides

from allennlp.common.util import JsonDict
from allennlp.data import Instance
from allennlp.predictors.predictor import Predictor

class BidafPredictor(Predictor):
    u"""
    Predictor for the :class:`~allennlp.models.bidaf.BidirectionalAttentionFlow` model.
    """

    def predict(self, question     , passage     )            :
        u"""
        Make a machine comprehension prediction on the supplied input.
        See https://rajpurkar.github.io/SQuAD-explorer/ for more information about the machine comprehension task.

        Parameters
        ----------
        question : ``str``
            A question about the content in the supplied paragraph.  The question must be answerable by a
            span in the paragraph.
        passage : ``str``
            A paragraph of information relevant to the question.

        Returns
        -------
        A dictionary that represents the prediction made by the system.  The answer string will be under the
        "best_span_str" key.
        """
        return self.predict_json({u"passage" : passage, u"question" : question})

    #overrides
    def _json_to_instance(self, json_dict          )            :
        u"""
        Expects JSON that looks like ``{"question": "...", "passage": "..."}``.
        """
        question_text = json_dict[u"question"]
        passage_text = json_dict[u"passage"]
        return self._dataset_reader.text_to_instance(question_text, passage_text)

BidafPredictor = Predictor.register(u'machine-comprehension')(BidafPredictor)



# pylint: disable=no-self-use,invalid-name


from __future__ import division
from __future__ import absolute_import
from allennlp.common.testing import AllenNlpTestCase
from allennlp.data.dataset_readers import NlvrDatasetReader
from allennlp.semparse.worlds import NlvrWorld


class TestNlvrDatasetReader(AllenNlpTestCase):
    def test_reader_reads_ungrouped_data(self):
        test_file = unicode(self.FIXTURES_ROOT / u"data" / u"nlvr" /
                        u"sample_ungrouped_data.jsonl")
        dataset = NlvrDatasetReader().read(test_file)
        instances = list(dataset)
        assert len(instances) == 3
        instance = instances[0]
        assert list(instance.fields.keys()) == set([u'sentence', u'agenda', u'worlds', u'actions', u'labels',
                                          u'identifier'])
        sentence_tokens = instance.fields[u"sentence"].tokens
        expected_tokens = [u'There', u'is', u'a', u'circle', u'closely', u'touching', u'a', u'corner', u'of',
                           u'a', u'box', u'.']
        assert [t.text for t in sentence_tokens] == expected_tokens
        actions = [action.rule for action in instance.fields[u"actions"].field_list]
        assert len(actions) == 115
        agenda = [item.sequence_index for item in instance.fields[u"agenda"].field_list]
        agenda_strings = [actions[rule_id] for rule_id in agenda]
        assert set(agenda_strings) == set([u'<o,o> -> circle',
                                           u'<o,t> -> object_exists',
                                           u'<o,o> -> touch_corner'])
        worlds = [world_field.as_tensor({}) for world_field in instance.fields[u"worlds"].field_list]
        assert isinstance(worlds[0], NlvrWorld)
        label = instance.fields[u"labels"].field_list[0].label
        assert label == u"true"

    def test_agenda_indices_are_correct(self):
        reader = NlvrDatasetReader()
        test_file = unicode(self.FIXTURES_ROOT / u"data" / u"nlvr" /
                        u"sample_ungrouped_data.jsonl")
        dataset = reader.read(test_file)
        instances = list(dataset)
        instance = instances[0]
        sentence_tokens = instance.fields[u"sentence"].tokens
        sentence = u" ".join([t.text for t in sentence_tokens])
        agenda = [item.sequence_index for item in instance.fields[u"agenda"].field_list]
        actions = [action.rule for action in instance.fields[u"actions"].field_list]
        agenda_actions = [actions[i] for i in agenda]
        world = instance.fields[u"worlds"].field_list[0].as_tensor({})
        expected_agenda_actions = world.get_agenda_for_sentence(sentence, add_paths_to_agenda=False)
        assert expected_agenda_actions == agenda_actions

    def test_reader_reads_grouped_data(self):
        test_file = unicode(self.FIXTURES_ROOT / u"data" / u"nlvr" /
                        u"sample_grouped_data.jsonl")
        dataset = NlvrDatasetReader().read(test_file)
        instances = list(dataset)
        assert len(instances) == 2
        instance = instances[0]
        assert list(instance.fields.keys()) == set([u'sentence', u'agenda', u'worlds', u'actions', u'labels',
                                          u'identifier'])
        sentence_tokens = instance.fields[u"sentence"].tokens
        expected_tokens = [u'There', u'is', u'a', u'circle', u'closely', u'touching', u'a', u'corner', u'of',
                           u'a', u'box', u'.']
        assert [t.text for t in sentence_tokens] == expected_tokens
        actions = [action.rule for action in instance.fields[u"actions"].field_list]
        assert len(actions) == 115
        agenda = [item.sequence_index for item in instance.fields[u"agenda"].field_list]
        agenda_strings = [actions[rule_id] for rule_id in agenda]
        assert set(agenda_strings) == set([u'<o,o> -> circle',
                                           u'<o,o> -> touch_corner',
                                           u'<o,t> -> object_exists'
                                          ])
        worlds = [world_field.as_tensor({}) for world_field in instance.fields[u"worlds"].field_list]
        assert all([isinstance(world, NlvrWorld) for world in worlds])
        labels = [label.label for label in instance.fields[u"labels"].field_list]
        assert labels == [u"true", u"false", u"true", u"false"]

    def test_reader_reads_processed_data(self):
        # Processed data contains action sequences that yield the correct denotations, obtained from
        # an offline search.
        test_file = unicode(self.FIXTURES_ROOT / u"data" / u"nlvr" /
                        u"sample_processed_data.jsonl")
        dataset = NlvrDatasetReader().read(test_file)
        instances = list(dataset)
        assert len(instances) == 2
        instance = instances[0]
        assert list(instance.fields.keys()) == set([u"sentence", u"target_action_sequences",
                                          u"worlds", u"actions", u"labels", u"identifier"])
        all_action_sequence_indices = instance.fields[u"target_action_sequences"].field_list
        assert len(all_action_sequence_indices) == 20
        action_sequence_indices = [item.sequence_index for item in
                                   all_action_sequence_indices[0].field_list]
        actions = [action.rule for action in instance.fields[u"actions"].field_list]
        action_sequence = [actions[rule_id] for rule_id in action_sequence_indices]
        assert action_sequence == [u'@start@ -> t',
                                   u't -> [<o,t>, o]',
                                   u'<o,t> -> object_exists',
                                   u'o -> [<o,o>, o]',
                                   u'<o,o> -> touch_corner',
                                   u'o -> [<o,o>, o]',
                                   u'<o,o> -> circle',
                                   u'o -> all_objects']



import hashlib
import numpy as np
import os

from itertools import islice, chain
try:
    from itertools import imap
except ImportError:
    imap = map
try:
    xrange
except NameError:
    xrange = range


DEFAULT_PRECISION = 7
DEFAULT_NGRAM_BEG = 3
DEFAULT_NGRAM_END = 6
SQLITE_TOKEN_SPLITTERS = ''.join([chr(i) for i in range(128)
                                  if not chr(i).isalnum()])
BOW = u"\uF000"
EOW = u"\uF000"
CONVERTER_VERSION = 2


# MD5s a file in chunks in a streaming fashion
def md5_file(path, block_size=256 * 128):
    md5 = hashlib.md5()
    with open(path, 'rb') as f:
        for chunk in iter(lambda: f.read(block_size), b''):
            md5.update(chunk)
    return md5.hexdigest()


def fast_md5_file(path, block_size=256 * 128, stream=False):
    if stream:
        return hashlib.md5(
            ('###MAGNITUDEURI:' +
             path).encode('utf-8')).hexdigest()
    else:
        md5 = hashlib.md5()
        f_size = os.path.getsize(path)
        if f_size <= 104857600:
            return md5_file(path, block_size)
        clipped_f_size = f_size - (block_size + 1)
        md5.update(str(f_size).encode('utf-8'))
        interval = 25
        seek_interval = int(float(clipped_f_size) / float(interval))
        with open(path, 'rb') as f:
            for i in range(interval):
                f.seek((i * seek_interval) % clipped_f_size)
                chunk = f.read(block_size)
                md5.update(chunk)
        return md5.hexdigest()


def char_ngrams(key, beg, end):
    return chain.from_iterable((imap(lambda ngram: ''.join(ngram), zip(
        *[key[i:] for i in xrange(j)])) for j in xrange(beg, min(len(key) + 1, end + 1))))  # noqa


def norm_matrix(m):
    return m / np.linalg.norm(m, axis=1).reshape((-1, 1))


def norm_elmo(e):
    for i in range(e.shape[0]):
        e[i, :, :] = norm_matrix(e[i, :, :])


def ibatch(iterable, size):
    sourceiter = iter(iterable)
    while True:
        batchiter = islice(sourceiter, size)
        yield chain([next(batchiter)], batchiter)



u"""
Reading comprehension is loosely defined as follows: given a question and a passage of text that
contains the answer, answer the question.

These submodules contain readers for things that are predominantly reading comprehension datasets.
"""


from __future__ import absolute_import
from allennlp.data.dataset_readers.reading_comprehension.squad import SquadReader
from allennlp.data.dataset_readers.reading_comprehension.triviaqa import TriviaQaReader



u"""
This module contains code for transition-based decoding.  "Transition-based decoding" is where you
start in some state, iteratively transition between states, and have some kind of supervision
signal that tells you which end states, or which transition sequences, are "good".

If you want to do decoding for a vocabulary-based model, where the allowable outputs are the same
at every timestep of decoding, this code is not what you are looking for, and it will be quite
inefficient compared to other things you could do.

The key abstractions in this code are the following:

    - ``DecoderState`` represents the current state of decoding, containing a list of all of the
      actions taken so far, and a current score for the state.  It also has methods around
      determining whether the state is "finished" and for combining states for batched computation.
    - ``DecoderStep`` is a ``torch.nn.Module`` that models the transition function between states.
      Its main method is ``take_step``, which generates a ranked list of next states given a
      current state.
    - ``DecoderTrainer`` is an algorithm for training the transition function with some kind of
      supervision signal.  There are many options for training algorithms and supervision signals;
      this is an abstract class that is generic over the type of the supervision signal.

The module also has some classes to help represent the ``DecoderState``, including ``RnnState``,
which you can use to keep track of a decoder RNN's internal state, ``GrammarState``, which
keeps track of what actions are allowed at each timestep of decoding, if your outputs are
production rules from a grammar, and ``ChecklistState`` that keeps track of coverage inforation if
you are training a coverage based parser.

There is also a generic ``BeamSearch`` class for finding the ``k`` highest-scoring transition
sequences given a trained ``DecoderStep`` and an initial ``DecoderState``.
"""

from __future__ import absolute_import
from allennlp.nn.decoding.beam_search import BeamSearch
from allennlp.nn.decoding.checklist_state import ChecklistState
from allennlp.nn.decoding.constrained_beam_search import ConstrainedBeamSearch
from allennlp.nn.decoding.decoder_state import DecoderState
from allennlp.nn.decoding.decoder_step import DecoderStep
from allennlp.nn.decoding.decoder_trainers.decoder_trainer import DecoderTrainer
from allennlp.nn.decoding.grammar_state import GrammarState
from allennlp.nn.decoding.rnn_state import RnnState




from __future__ import absolute_import
from copy import deepcopy
#typing


class GrammarState(object):
    u"""
    A ``GrammarState`` specifies the currently valid actions at every step of decoding.

    If we had a global context-free grammar, this would not be necessary - the currently valid
    actions would always be the same, and we would not need to represent the current state.
    However, our grammar is not context free (we have lambda expressions that introduce
    context-dependent production rules), and it is not global (each instance can have its own
    entities of a particular type, or its own functions).

    We thus recognize three different sources of valid actions.  The first are actions that come
    from the type declaration; these are defined once by the model and shared across all
    ``GrammarStates`` produced by that model.  The second are actions that come from the current
    instance; these are defined by the ``World`` that corresponds to each instance, and are shared
    across all decoding states for that instance.  The last are actions that come from the current
    state of the decoder; these are updated after every action taken by the decoder, though only
    some actions initiate changes.

    In practice, we use the ``World`` class to get the first two sources of valid actions at the
    same time, and we take as input a ``valid_actions`` dictionary that is computed by the
    ``World``.  These will not change during the course of decoding.  The ``GrammarState`` object
    itself maintains the context-dependent valid actions.

    Parameters
    ----------
    nonterminal_stack : ``List[str]``
        Holds the list of non-terminals that still need to be expanded.  This starts out as
        [START_SYMBOL], and decoding ends when this is empty.  Every time we take an action, we
        update the non-terminal stack and the context-dependent valid actions, and we use what's on
        the stack to decide which actions are valid in the current state.
    lambda_stacks : ``Dict[Tuple[str, str], List[str]]``
        The lambda stack keeps track of when we're in the scope of a lambda function.  The
        dictionary is keyed by the production rule we are adding (like "r -> x", separated into
        left hand side and right hand side, where the LHS is the type of the lambda variable and
        the RHS is the variable itself), and the value is a nonterminal stack much like
        ``nonterminal_stack``.  When the stack becomes empty, we remove the lambda entry.
    valid_actions : ``Dict[str, List[int]]``
        A mapping from non-terminals (represented as strings) to all valid (global and
        instance-specific) productions from that non-terminal (represented as a list of integers).
    action_indices : ``Dict[str, int]``
        We use integers to represent productions in the ``valid_actions`` dictionary for efficiency
        reasons in the decoder.  This means we need a way to map from the production rule strings
        that we generate for lambda variables back to the integer used to represent it.
    is_nonterminal : ``Callable[[str], bool]``
        A function that is used to determine whether each piece of the RHS of the action string is
        a non-terminal that needs to be added to the non-terminal stack.  You can use
        ``type_declaraction.is_nonterminal`` here, or write your own function if that one doesn't
        work for your domain.
    """
    def __init__(self,
                 nonterminal_stack           ,
                 lambda_stacks                                  ,
                 valid_actions                      ,
                 action_indices                ,
                 is_nonterminal                       )        :
        self._nonterminal_stack = nonterminal_stack
        self._lambda_stacks = lambda_stacks
        self._valid_actions = valid_actions
        self._action_indices = action_indices
        self._is_nonterminal = is_nonterminal

    def is_finished(self)        :
        u"""
        Have we finished producing our logical form?  We have finished producing the logical form
        if and only if there are no more non-terminals on the stack.
        """
        return not self._nonterminal_stack

    def get_valid_actions(self)             :
        u"""
        Returns a list of valid actions (represented as integers)
        """
        actions = self._valid_actions[self._nonterminal_stack[-1]]
        for type_, variable in self._lambda_stacks:
            if self._nonterminal_stack[-1] == type_:
                production_string = "{type_} -> {variable}"
                actions = actions + [self._action_indices[production_string]]
        return actions

    def take_action(self, production_rule     )                  :
        u"""
        Takes an action in the current grammar state, returning a new grammar state with whatever
        updates are necessary.  The production rule is assumed to be formatted as "LHS -> RHS".

        This will update the non-terminal stack and the context-dependent actions.  Updating the
        non-terminal stack involves popping the non-terminal that was expanded off of the stack,
        then pushing on any non-terminals in the production rule back on the stack.  We push the
        non-terminals on in `reverse` order, so that the first non-terminal in the production rule
        gets popped off the stack first.

        For example, if our current ``nonterminal_stack`` is ``["r", "<e,r>", "d"]``, and
        ``action`` is ``d -> [<e,d>, e]``, the resulting stack will be ``["r", "<e,r>", "e",
        "<e,d>"]``.
        """
        left_side, right_side = production_rule.split(u' -> ')
        assert self._nonterminal_stack[-1] == left_side, ("Tried to expand {self._nonterminal_stack[-1]}"
                                                          u"but got rule f{left_side}->f{right_side}")
        new_stack = self._nonterminal_stack[:-1]
        new_lambda_stacks = deepcopy(self._lambda_stacks)
        for key, lambda_stack in list(new_lambda_stacks.items()):
            assert lambda_stack[-1] == left_side
            lambda_stack.pop()  # pop to modify the value in the dictionary

        productions = self._get_productions_from_string(right_side)
        # Looking for lambda productions, but not for cells or columns with the word "lambda" in
        # them.
        if u'lambda' in productions[0] and u'fb:' not in productions[0]:
            production = productions[0]
            if production[0] == u"'" and production[-1] == u"'":
                # The production rule with a lambda is typically "<t,d> -> ['lambda x', d]".  We
                # need to strip the quotes.
                production = production[1:-1]
            lambda_variable = production.split(u' ')[1]
            # The left side must be formatted as "<t,d>", where "t" is the type of the lambda
            # variable, and "d" is the return type of the lambda function.  We need to pull out the
            # "t" here.  TODO(mattg): this is pretty limiting, but I'm not sure how general we
            # should make this.
            if len(left_side) != 5:
                raise NotImplementedError(u"Can't handle this type yet:", left_side)
            lambda_type = left_side[1]
            new_lambda_stacks[(lambda_type, lambda_variable)] = []

        for production in reversed(productions):
            if self._is_nonterminal(production):
                new_stack.append(production)
                for lambda_stack in list(new_lambda_stacks.values()):
                    lambda_stack.append(production)

        # If any of the lambda stacks have now become empty, we remove them from our dictionary.
        finished_lambdas = set()
        for key, lambda_stack in list(new_lambda_stacks.items()):
            if not lambda_stack:
                finished_lambdas.add(key)
        for finished_lambda in finished_lambdas:
            del new_lambda_stacks[finished_lambda]

        return GrammarState(nonterminal_stack=new_stack,
                            lambda_stacks=new_lambda_stacks,
                            valid_actions=self._valid_actions,
                            action_indices=self._action_indices,
                            is_nonterminal=self._is_nonterminal)

    @staticmethod
    def _get_productions_from_string(production_string     )             :
        u"""
        Takes a string like '[<d,d>, d]' and parses it into a list like ['<d,d>', 'd'].  For
        production strings that are not lists, like '<e,d>', we return a single-element list:
        ['<e,d>'].
        """
        if production_string[0] == u'[':
            return production_string[1:-1].split(u', ')
        else:
            return [production_string]



# pylint: disable=no-self-use,invalid-name,too-many-public-methods
from __future__ import absolute_import
#typing

from allennlp.common import Params
from allennlp.common.from_params import FromParams, takes_arg, remove_optional, create_kwargs
from allennlp.common.testing import AllenNlpTestCase
from allennlp.data.tokenizers.word_splitter import WordSplitter

class MyClass(FromParams):
    def __init__(self, my_int     , my_bool       = False)        :
        self.my_int = my_int
        self.my_bool = my_bool


class TestFromParams(AllenNlpTestCase):
    def test_takes_arg(self):
        def bare_function(some_input     )       :
            return some_input + 1

        assert takes_arg(bare_function, u'some_input')
        assert not takes_arg(bare_function, u'some_other_input')

        class SomeClass(object):
            total = 0

            def __init__(self, constructor_param     )        :
                self.constructor_param = constructor_param

            def check_param(self, check     )        :
                return self.constructor_param == check

            @classmethod
            def set_total(cls, new_total     )        :
                cls.total = new_total

        assert takes_arg(SomeClass, u'self')
        assert takes_arg(SomeClass, u'constructor_param')
        assert not takes_arg(SomeClass, u'check')

        assert takes_arg(SomeClass.check_param, u'check')
        assert not takes_arg(SomeClass.check_param, u'other_check')

        assert takes_arg(SomeClass.set_total, u'new_total')
        assert not takes_arg(SomeClass.set_total, u'total')

    def test_remove_optional(self):
        optional_type = Optional[Dict[unicode, unicode]]
        bare_type = remove_optional(optional_type)
        bare_bare_type = remove_optional(bare_type)

        assert bare_type == Dict[unicode, unicode]
        assert bare_bare_type == Dict[unicode, unicode]

        assert remove_optional(Optional[unicode]) == unicode
        assert remove_optional(unicode) == unicode

    def test_from_params(self):
        my_class = MyClass.from_params(Params({u"my_int": 10}), my_bool=True)

        assert isinstance(my_class, MyClass)
        assert my_class.my_int == 10
        assert my_class.my_bool

    def test_create_kwargs(self):
        kwargs = create_kwargs(MyClass,
                               Params({u'my_int': 5}),
                               my_bool=True,
                               my_float=4.4)

        # my_float should not be included because it's not a param of the MyClass constructor
        assert kwargs == {
                u"my_int": 5,
                u"my_bool": True
        }

    def test_extras(self):
        # pylint: disable=unused-variable,arguments-differ
        from allennlp.common.registrable import Registrable

        class A(Registrable):
            pass

        class B(A):
            def __init__(self, size     , name     )        :
                self.size = size
                self.name = name

        B = A.register(u"b")(B)

class C(A):
            def __init__(self, size     , name     )        :
                self.size = size
                self.name = name

            # custom from params
            @classmethod
            def from_params(cls, params        , size     )       :  # type: ignore
                name = params.pop(u'name')
                return cls(size=size, name=name)


        # Check that extras get passed, even though A doesn't need them.
        C = A.register(u"c")(C)

params = Params({u"type": u"b", u"size": 10})
        b = A.from_params(params, name=u"extra")

        assert b.name == u"extra"
        assert b.size == 10

        # Check that extra extras don't get passed.
        params = Params({u"type": u"b", u"size": 10})
        b = A.from_params(params, name=u"extra", unwanted=True)

        assert b.name == u"extra"
        assert b.size == 10

        # Now the same with a custom from_params.
        params = Params({u"type": u"c", u"name": u"extra_c"})
        c = A.from_params(params, size=20)
        assert c.name == u"extra_c"
        assert c.size == 20

        # Check that extra extras don't get passed.
        params = Params({u"type": u"c", u"name": u"extra_c"})
        c = A.from_params(params, size=20, unwanted=True)

        assert c.name == u"extra_c"
        assert c.size == 20

    def test_no_constructor(self):
        params = Params({u"type": u"just_spaces"})

        WordSplitter.from_params(params)



u"""
Reading comprehension is loosely defined as follows: given a question and a passage of text that
contains the answer, answer the question.

These submodules contain models for things that are predominantly focused on reading comprehension.
"""


from __future__ import absolute_import
from allennlp.models.reading_comprehension.bidaf import BidirectionalAttentionFlow
from allennlp.models.reading_comprehension.bidaf_ensemble import BidafEnsemble




from __future__ import absolute_import
import torch

from allennlp.common import Registrable

class Regularizer(Registrable):
    u"""
    An abstract class representing a regularizer. It must implement
    call, returning a scalar tensor.
    """
    default_implementation = u'l2'

    def __call__(self, parameter              )                :
        raise NotImplementedError



# pylint: disable=no-self-use,invalid-name
from __future__ import with_statement
from __future__ import division
from __future__ import absolute_import
import pytest

from allennlp.common.testing import ModelTestCase
from allennlp.data.dataset import Batch
from io import open

# Skip this one, it's an expensive test.
class TestOpenaiTransformerEmbedderLarge(ModelTestCase):
    def setUp(self):
        super(TestOpenaiTransformerEmbedderLarge, self).setUp()
        self.set_up_model(self.FIXTURES_ROOT / u'openai_transformer' / u'config_large.jsonnet',
                          self.FIXTURES_ROOT / u'data' / u'conll2003.txt')

    def test_tagger_with_openai_token_embedder_can_train_save_and_load(self):
        self.ensure_model_can_train_save_and_load(self.param_file)

    def test_tagger_with_openai_token_embedder_forward_pass_runs_correctly(self):
        dataset = Batch(self.instances)
        dataset.index_instances(self.vocab)
        training_tensors = dataset.as_tensor_dict()
        output_dict = self.model(**training_tensors)
        tags = output_dict[u'tags']
        assert len(tags) == 2
        assert len(tags[0]) == 7
        assert len(tags[1]) == 7
        for example_tags in tags:
            for tag_id in example_tags:
                tag = self.model.vocab.get_token_from_index(tag_id, namespace=u"labels")
                assert tag in set([u'O', u'I-ORG', u'I-PER', u'I-LOC'])


TestOpenaiTransformerEmbedderLarge =  Skip this one, it's an expensive test.
@pytest.mark.skip()(TestOpenaiTransformerEmbedderLarge)

class TestOpenaiTransformerEmbedderSmall(ModelTestCase):
    def setUp(self):
        super(TestOpenaiTransformerEmbedderSmall, self).setUp()
        self.set_up_model(self.FIXTURES_ROOT / u'openai_transformer' / u'config_small.jsonnet',
                          self.FIXTURES_ROOT / u'data' / u'conll2003.txt')

    def test_tagger_with_openai_token_embedder_can_train_save_and_load(self):
        self.ensure_model_can_train_save_and_load(self.param_file)

    def test_tagger_with_openai_token_embedder_forward_pass_runs_correctly(self):
        dataset = Batch(self.instances)
        dataset.index_instances(self.vocab)
        training_tensors = dataset.as_tensor_dict()
        output_dict = self.model(**training_tensors)
        tags = output_dict[u'tags']
        assert len(tags) == 2
        assert len(tags[0]) == 7
        assert len(tags[1]) == 7
        for example_tags in tags:
            for tag_id in example_tags:
                tag = self.model.vocab.get_token_from_index(tag_id, namespace=u"labels")
                assert tag in set([u'O', u'I-ORG', u'I-PER', u'I-LOC'])


def create_small_test_fixture(output_dir      = u'/tmp')        :
    u"""
    This is how I created the transformer_model.tar.gz.
    After running this, go to the specified output dir and run

        tar -czvf transformer_model.tar.gz model/

    In case you need to regenerate the fixture for some reason.
    """
    import json
    import pathlib
    from allennlp.modules.openai_transformer import OpenaiTransformer

    model_dir = pathlib.Path(output_dir) / u'model'
    model_dir.mkdir(exist_ok=True)  # pylint: disable=no-member

    symbols = [u"e", u"w", u"o", u"wo", u"."]
    byte_pairs = [(sym1, sym2 + end)
                  for sym1 in symbols        # prefer earlier first symbol
                  for sym2 in symbols        # if tie, prefer earlier second symbol
                  for end in (u'</w>', u'')]   # if tie, prefer ending a word
    encoding = dict(("{sym1}{sym2}", idx) for idx, (sym1, sym2) in enumerate(byte_pairs))
    encoding[u"<unk>"] = 0

    with open(model_dir / u'encoder_bpe.json', u'w') as encoder_file:
        json.dump(encoding, encoder_file)

    with open(model_dir / u'vocab.bpe', u'w') as bpe_file:
        bpe_file.write(u"#version 0.0\n")
        for sym1, sym2 in byte_pairs:
            bpe_file.write("{sym1} {sym2}\n")
        bpe_file.write(u"\n")

    transformer = OpenaiTransformer(embedding_dim=10, num_heads=2, num_layers=2, vocab_size=(50 + 50), n_ctx=50)
    transformer.dump_weights(output_dir, num_pieces=2)



# pylint: disable=invalid-name,no-self-use



from __future__ import with_statement
from __future__ import division
from __future__ import absolute_import
import argparse
import re
import shutil

import pytest

from allennlp.common.testing import AllenNlpTestCase
from allennlp.commands.fine_tune import FineTune, fine_tune_model_from_file_paths,\
                               fine_tune_model_from_args, fine_tune_model
from allennlp.common.params import Params
from allennlp.models import load_archive

class TestFineTune(AllenNlpTestCase):
    def setUp(self):
        super(TestFineTune, self).setUp()
        self.model_archive = unicode(self.FIXTURES_ROOT / u'decomposable_attention' / u'serialization' / u'model.tar.gz')
        self.config_file = unicode(self.FIXTURES_ROOT / u'decomposable_attention' / u'experiment.json')
        self.serialization_dir = unicode(self.TEST_DIR / u'fine_tune')

        self.parser = argparse.ArgumentParser(description=u"Testing")
        subparsers = self.parser.add_subparsers(title=u'Commands', metavar=u'')
        FineTune().add_subparser(u'fine-tune', subparsers)


    def test_fine_tune_model_runs_from_file_paths(self):
        fine_tune_model_from_file_paths(model_archive_path=self.model_archive,
                                        config_file=self.config_file,
                                        serialization_dir=self.serialization_dir)

    def test_fine_tune_does_not_expand_vocab_by_default(self):
        params = Params.from_file(self.config_file)
        # snli2 has a new token in it
        params[u"train_data_path"] = unicode(self.FIXTURES_ROOT / u'data' / u'snli2.jsonl')

        model = load_archive(self.model_archive).model

        # By default, no vocab expansion.
        fine_tune_model(model, params, self.serialization_dir)

    def test_fine_tune_runtime_errors_with_vocab_expansion(self):
        params = Params.from_file(self.config_file)
        params[u"train_data_path"] = unicode(self.FIXTURES_ROOT / u'data' / u'snli2.jsonl')

        model = load_archive(self.model_archive).model

        # If we do vocab expansion, we get a runtime error because of the embedding.
        with pytest.raises(RuntimeError):
            fine_tune_model(model, params, self.serialization_dir, extend_vocab=True)

    def test_fine_tune_runs_from_parser_arguments(self):
        raw_args = [u"fine-tune",
                    u"-m", self.model_archive,
                    u"-c", self.config_file,
                    u"-s", self.serialization_dir]

        args = self.parser.parse_args(raw_args)

        assert args.func == fine_tune_model_from_args
        assert args.model_archive == self.model_archive
        assert args.config_file == self.config_file
        assert args.serialization_dir == self.serialization_dir
        fine_tune_model_from_args(args)

    def test_fine_tune_fails_without_required_args(self):
        # Configuration file is required.
        with self.assertRaises(SystemExit) as context:
            self.parser.parse_args([u"fine-tune", u"-m", u"path/to/archive", u"-s", u"serialization_dir"])
            assert context.exception.code == 2  # argparse code for incorrect usage

        # Serialization dir is required.
        with self.assertRaises(SystemExit) as context:
            self.parser.parse_args([u"fine-tune", u"-m", u"path/to/archive", u"-c", u"path/to/config"])
            assert context.exception.code == 2  # argparse code for incorrect usage

        # Model archive is required.
        with self.assertRaises(SystemExit) as context:
            self.parser.parse_args([u"fine-tune", u"-s", u"serialization_dir", u"-c", u"path/to/config"])
            assert context.exception.code == 2  # argparse code for incorrect usage

    def test_fine_tune_nograd_regex(self):
        original_model = load_archive(self.model_archive).model
        name_parameters_original = dict(original_model.named_parameters())
        regex_lists = [[],
                       [u".*attend_feedforward.*", u".*token_embedder.*"],
                       [u".*compare_feedforward.*"]]
        for regex_list in regex_lists:
            params = Params.from_file(self.config_file)
            params[u"trainer"][u"no_grad"] = regex_list
            shutil.rmtree(self.serialization_dir, ignore_errors=True)
            tuned_model = fine_tune_model(model=original_model,
                                          params=params,
                                          serialization_dir=self.serialization_dir)
            # If regex is matched, parameter name should have requires_grad False
            # If regex is matched, parameter name should have same requires_grad
            # as the originally loaded model
            for name, parameter in tuned_model.named_parameters():
                if any(re.search(regex, name) for regex in regex_list):
                    assert not parameter.requires_grad
                else:
                    assert parameter.requires_grad\
                    == name_parameters_original[name].requires_grad
        # If all parameters have requires_grad=False, then error.
        with pytest.raises(Exception) as _:
            params = Params.from_file(self.config_file)
            params[u"trainer"][u"no_grad"] = [u"*"]
            shutil.rmtree(self.serialization_dir, ignore_errors=True)
            tuned_model = fine_tune_model(model=original_model,
                                          params=params,
                                          serialization_dir=self.serialization_dir)
